package web_server.sim.schelling_180206;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.UUID;

import processing.core.PVector;
import processing.data.JSONArray;
import processing.data.JSONObject;

public class Agent_Human extends Agent {
	ArrayList<Agent_Venue> venues;								// reference from the other rule..
	public Agent[][] occupiedBy;

	int worldSize;												// stored locally: the grid size of the simulation world 
	int numCultures = 0;										// fixed to 2 for this version										
	String movement;												// best, closest, random that works
	boolean vonNeumann;											// whether a von neumann or moore neighborhood
	String learning;										// random, inertial, direction
	int maxTime;
	int cultureIndex = -1;
	
	float learnOblig = 0;										// if learning type is directional, these store the direction.
	float learnExclu = 0;
	
	ArrayList<Agent_Venue> venuesVisited = new ArrayList<Agent_Venue>();	// stores a list of the venues visited in a given turn
	ArrayList<Agent_Venue> venuesNotVisited = new ArrayList<Agent_Venue>();	// stores a list of the venues not visited, but that could have been based on distance
	
	float rndOblig = 0;						// these values are inertial randoms - they are intended to change slowly over time
	float rndExclu = 0;						// but also to be able to become more rigid
	public boolean unset = false;			// unsettled, i.e. it wants to move but couldn't find a reasonable destination
	public boolean moved = false;			// moved successfully this time step
	
	double randomMove = 0;
	// things to store each time step
	// moved : yes/no
	// position
	// oblig, exclu
	// uset
	// list of venues (uid) that were visited
	float[] store_rndOblig;
	float[] store_rndExclu;
	boolean[] store_unset;
	boolean[] store_moved;
	int[][] store_pos;
	ArrayList<String>[] store_visited;
	
	public Agent_Human (int worldSize, int numCultures, Culture cult, String mType, ArrayList<Agent_Venue> venues, Agent[][] occupiedBy, java.util.Random rnd, boolean vn, String learningType, int maxTime, PVector pos, double rndMove) {
		this.uid = UUID.randomUUID();
		this.type = "human";
		this.worldSize = worldSize;
		this.numCultures = numCultures;
		this.venues = venues;
		this.occupiedBy = occupiedBy;
		this.vonNeumann = vn;
		this.learning = learningType;
		this.maxTime = maxTime;
		this.randomMove = rndMove;
		
		this.cult = cult;
		this.cultureIndex = cult.identityCode;
		this.movement = mType;
		this.rndOblig = (float) rnd.nextDouble();
		this.rndExclu = (float) rnd.nextDouble();		

		store_rndOblig = new float[maxTime+1];
		store_rndExclu = new float[maxTime+1];
		store_unset = new boolean[maxTime+1];
		store_moved = new boolean[maxTime+1];
		store_pos = new int[maxTime+1][];
		store_visited = new ArrayList[maxTime+1];
		
		if (this.learning.equals("Direction")) {
			learnOblig = 0;
			learnExclu = 0;
		}
		
		// if no position was provided
		if (pos == null)
			FindAPos( rnd );
		else {
			if (occupiedBy[(int) pos.x][(int) pos.y] == null) {
				this.pos = new int[] { (int) pos.x, (int) pos.y };
				occupiedBy[this.pos[0]][this.pos[1]] = this;
			} else {
				// if the spot is already occupied, we'll find a location for this agent later.
				this.pos = null;
			}
		}
	}
	
	
	public Agent_Human(JSONObject haa) {
		this.type = "human";		
		this.uid = UUID.fromString(haa.getString("id"));
		this.store_rndOblig = haa.getJSONArray("Oblig").getFloatArray();
		this.store_rndExclu = haa.getJSONArray("Exclu").getFloatArray();
		this.store_moved = haa.getJSONArray("Moved").getBooleanArray();
		this.learning = haa.getString("Learning");
		this.store_unset = haa.getJSONArray("Unset").getBooleanArray();
		this.movement = haa.getString("Movement");
		this.cultureIndex = haa.getInt("Culture");
		
		JSONArray jsp = haa.getJSONArray("Pos");
		this.store_pos = new int[jsp.size()][];
		
		for (int i = 0; i < store_pos.length; i++) {
			String[] sd = jsp.getString(i).split(" ");
			store_pos[i] = new int[] { Integer.parseInt(sd[0]), Integer.parseInt(sd[1]) };
		}
		JSONArray jv = haa.getJSONArray("Visit");
		this.store_visited = new ArrayList[jv.size()];
		
		for (int i = 0; i < store_visited.length; i++) {
			store_visited[i] =  new ArrayList<String>(Arrays.asList( jv.getJSONArray(i).getStringArray() ));
		}
	}

	public void FindAPos( java.util.Random rnd ) {
		int breakOut = 0;
		boolean finished = false;
		
		do {
			int x = (int) (rnd.nextDouble() * worldSize);
			int y = (int) (rnd.nextDouble() * worldSize);
			
			if (occupiedBy[x][y] == null) {
				pos = new int[] { x, y };
				occupiedBy[x][y] = this;
				finished = true;
			}
		} while (!finished && breakOut < worldSize*worldSize*worldSize);
		
		if (pos==null) {
			System.out.println("error, unable to place an agent - too dense?");
		}
	}
	
	public void GoToVenues( int t ) {
		venuesVisited.clear();
		venuesNotVisited.clear();
		
		for (Agent_Venue v : this.venues) {
			// if it's in range
			if ( InRange(v, this.pos[0], this.pos[1]) ) {
				boolean vi = false;
				
				if ( this.rndOblig > 1 - v.oblig ) {
					if ( v.cult.identityCode == this.cult.identityCode || this.rndExclu > v.exclu ) {

						// it visited the venue...
						vi = true;
						v.visitors.add( this );
						this.venuesVisited.add(v);
					}
				}
				
				if (!vi) {
					this.venuesNotVisited.add(v);
				}
			}
		}		
	}
	
	/* - - - - - - - - - - - - - - - - - */
	public void Update( int t ) {		
		unset = false;					// similar but different
		moved = false;
		
		if ( t > 0 && !ActuallyValid(this.pos[0], this.pos[1] ) ) {
			// if the agent is unhappy at their current position.
			
			occupiedBy[pos[0]][pos[1]] = null;
			
			int[] np = new int[] {0,0};
			if (movement.equals("Best")) {
				np = findBestValid();					// best
			} else if (movement.equals("Closest")) {
				np = findNearestValid();				// nearest
			} else if (movement.equals("Rand")){
				np = findRandomValid();					//random
			}
			
			try {
			    Files.write(Paths.get("moves.txt"), (pos[0] + " " + pos[1] + "," + np[0] + " " + np[1] + "," + this.cultureIndex + "\n").getBytes(), StandardOpenOption.APPEND);
			}catch (Exception e) {

			}
			
			
			pos = new int[] { np[0], np[1] };
			
			occupiedBy[pos[0]][pos[1]] = this;
		} else {
			
			if (t == 0) {
				try {
				    Files.write(Paths.get("moves.txt"), (pos[0] + " " + pos[1] + " " + this.cultureIndex + "\n").getBytes(), StandardOpenOption.APPEND);
				}catch (Exception e) {

				}
			}
			
			// time 0, or the agent is happy where they are
			if (t > 0 && Math.random() < randomMove) {
				occupiedBy[pos[0]][pos[1]] = null;
				
				int[] np = new int[] {0,0};
				if (movement.equals("Best")) {
					np = findBestValid();					// best
				} else if (movement.equals("Closest")) {
					np = findNearestValid();				// nearest
				} else if (movement.equals("Rand")){
					np = findRandomValid();					//random
				}			
				pos = new int[] { np[0], np[1] };
				occupiedBy[pos[0]][pos[1]] = this;
			}
		}
		Learn();
		StoreStep(t);
	}
	
	public void StoreStep(int t) {
		store_rndOblig[t] = this.rndOblig;
		store_rndExclu[t] = this.rndExclu;
		store_unset[t] = this.unset;
		store_moved[t] = this.moved;
		store_pos[t] = this.pos;
		store_visited[t] = new ArrayList<String>();

		for (Agent_Venue v : this.venuesVisited) {
			store_visited[t].add( v.uid.toString() );
		}
	}
	
	public JSONObject getJSON() {
		JSONObject jo = new JSONObject();
		jo.setString("id", this.uid.toString());
		jo.setInt("Culture", this.cult.identityCode);
		jo.setString("Movement", this.movement);
		jo.setString("Learning", this.learning);
		JSONArray joblig = new JSONArray();
		JSONArray jexclu = new JSONArray();
		JSONArray junset = new JSONArray();
		JSONArray jmoved = new JSONArray();
		JSONArray jpos = new JSONArray();
		JSONArray jvisit = new JSONArray();
		for (int i = 0; i < maxTime; i++) {
			joblig.append(store_rndOblig[i]);
			jexclu.append(store_rndExclu[i]);
			junset.append(store_unset[i]);
			jmoved.append(store_moved[i]);
			jpos.append(store_pos[i][0] + " " + store_pos[i][1]);
			JSONArray jv = new JSONArray();
			for (int j = 0; j < store_visited[i].size(); j++) {
				jv.append(store_visited[i].get(j));
			}
			jvisit.append(jv);
		}
		jo.setJSONArray("Oblig", joblig);
		jo.setJSONArray("Exclu", jexclu);
		jo.setJSONArray("Unset", junset);
		jo.setJSONArray("Moved", jmoved);
		jo.setJSONArray("Pos", jpos);
		jo.setJSONArray("Visit", jvisit);
		return jo;
	}

		
	public void Learn() {
		if (learning.equals("Random")) {
			// full random
			this.rndOblig = (float) Math.random();
			this.rndExclu = (float) Math.random();
		} else if (learning.equals("Inertial")) {
			// inertial
			
			for (Agent_Venue v : venuesVisited) {
				this.rndOblig += (float) Math.random() * 0.1f - 0.05f;
				this.rndExclu += (float) Math.random() * 0.1f - 0.05f;

				// the "1-" here, to acknowledge the inverse relationship
				this.rndOblig += (this.rndOblig - (1-v.oblig)  )/Math.abs(this.rndOblig - (1-v.oblig)) * 0.1;
				this.rndOblig = Math.min(1,  Math.max(0, this.rndOblig));

				this.rndExclu += (this.rndExclu - v.exclu)/Math.abs(this.rndExclu - v.exclu) * 0.1;
				this.rndExclu = Math.min(1, Math.max(0,  this.rndExclu));
			}
			for (Agent_Venue v : venuesNotVisited) {
				this.rndOblig += (float) Math.random() * 0.1f - 0.05f;
				this.rndExclu += (float) Math.random() * 0.1f - 0.05f;

				// the "1-" here, to acknowledge the inverse relationship
				this.rndOblig += (this.rndOblig - (1-v.oblig)  )/Math.abs(this.rndOblig - (1-v.oblig)) * 0.1;
				this.rndOblig = Math.min(1,  Math.max(0, this.rndOblig));

				this.rndExclu += (this.rndExclu - v.exclu)/Math.abs(this.rndExclu - v.exclu) * 0.1;
				this.rndExclu = Math.min(1, Math.max(0,  this.rndExclu));
			}
		} else if (learning.equals("Direction")) {
			// random direction
			this.rndOblig += learnOblig;
			this.rndExclu += learnExclu;
		}		
	}
	
	public boolean InRange(Agent_Venue v, int x, int y) {
		return (	this.vonNeumann && Math.sqrt((v.pos[0]-x)*(v.pos[0]-x) + (v.pos[1]-y)*(v.pos[1]-y) ) <= cult.venueTravelDist)
				|| (!this.vonNeumann && Math.max(Math.abs(v.pos[0]-x), Math.abs(v.pos[1]-y)) <= cult.venueTravelDist );
		
	}
	
	public int[] findRandomValid() {
		boolean finished = false;
		int maxAttempts = 2500;
		
		// I should really keep a list of empty cells!
		int att = 0;
		do {
			int x = (int) (Math.random() * worldSize);
			int y = (int) (Math.random() * worldSize);
			
			if (occupiedBy[x][y] == null && valid(x,y) ) {
				moved = true;
				return new int[] { x, y };
			}
			att++;
		} while (!finished && att < maxAttempts );		
		unset = true;
		return this.pos;
	}
	
	public int[] findBestValid ( ) {
		// loop through the whole world, keeping the best that has been found so far
		int x = this.pos[0];
		int y = this.pos[1];
		double[] bestScore = new double[] { 0, 0 };
		ArrayList<int[]> bestPositions = new ArrayList<int[]>();	// when a certain position is "best", we include all others that have the same score for 
																	// consideration, when a better score is found, these are all tossed out
		
		for (int i = 0; i < worldSize; i++) {
			for (int j = 0; j < worldSize; j++) {
				if (x!=i||y!=j) {
					
					if (occupiedBy[i][j] == null && valid(i,j) ) {
						double[] sc = getScore(i,j);
						
						if (sc[0] <= cult.maxCohesionParameter && sc[0] > bestScore[0]) {
							x = i; y = j;
							bestScore = sc;
							bestPositions.clear();
							bestPositions.add(new int[] {i,j} );
							
						} else if (sc[0] <= cult.maxCohesionParameter && sc[0] == bestScore[0] && sc[1] > bestScore[1]) {
							// same score percentage, but more individuals
							x = i; y = j;
							bestScore = sc;
							bestPositions.clear();
							bestPositions.add(new int[] {i,j} );
							
						} else if (sc[0] <= cult.maxCohesionParameter && sc[0] == bestScore[0] && sc[1] == bestScore[1]) {
							// same score percentage, same number of individuals
							bestPositions.add(new int[] {i,j} );
						}						
					}
				}
			}
		}
		// if more than one had the same best score and total num, randomly pick one
		if (bestPositions.size() > 1) {
			int[] nbp = bestPositions.get( (int) Math.random() * bestPositions.size() );
			x = nbp[0]; y = nbp[1];
		}
		
		if (bestScore[0] > 0)
			moved = true;
		else
			unset = true;
		return new int[] { x,y };
	}
	
	public int[] findNearestValid( ) {
		// loop through all of them, should spiral out randomly from the inital position.
		int x = this.pos[0];
		int y = this.pos[1];
		
		for (int stepDist = 1; stepDist < worldSize; stepDist++) {
			ArrayList<int[]> validPos = new ArrayList<int[]>();
			
			// moving on the x through the whole ring...
			for (int a = -stepDist; a < stepDist+1; a++) {
				
				// if we're on one of the vertical sides of the ring
				if (Math.abs(a)==stepDist) {
					for (int b = -stepDist; b < stepDist+1; b++) {
						if (x+a >= 0 && x+a < worldSize && y+b >= 0 && y+b < worldSize) {
							if (occupiedBy[x+a][y+b] == null && valid(x+a,y+b)) {
								validPos.add(new int[] { x+a, y+b });
							}
						}
					}
				} else {
					// jump between the two horizontal bars
					for (int b = -stepDist; b < stepDist+1; b += (stepDist*2)) {
						if (x+a >= 0 && x+a < worldSize && y+b >= 0 && y+b < worldSize) {
							if (occupiedBy[x+a][y+b] == null && valid(x+a,y+b)) {
								validPos.add(new int[] { x+a, y+b });
							}
						}
					}
				}
				
				if (validPos.size() > 0) {
					moved = true;
					return validPos.get( (int) (Math.random() * validPos.size()) );
				}
			}
		}
		
		unset = true;
		return new int[] {x,y};
	}
	
	// this is used to determine if the agent would like the square it's considering
	public boolean valid(int x, int y) {
		int[] neighbCount = new int[ numCultures ];
		int totalNeigh = 0;
		
		totalNeigh += GetInfluenceOfAgents(neighbCount,x,y);
		totalNeigh += GetInfluenceOfVenues(neighbCount,x,y);
		
		float neiVal = (neighbCount[this.cult.identityCode] / (float) totalNeigh);
		return totalNeigh == 0 || (neiVal >= cult.cohesionParameter && neiVal <= cult.maxCohesionParameter);
	}
	
	// this is used to determine if the agent would like the square it currently occupies - slightly different from above, because
	// this is based on what venues the agent has visited this turn
	public boolean ActuallyValid(int x, int y) {
		int[] neighbCount = new int[ numCultures ];
		int totalNeigh = 0;
		
		totalNeigh += GetInfluenceOfAgents(neighbCount,x,y);
		totalNeigh += GetInfluenceOfVenuesVisited(neighbCount);
		
		float neiVal = (neighbCount[this.cult.identityCode] / (float) totalNeigh);
		return totalNeigh == 0 || (neiVal >= cult.cohesionParameter && neiVal <= cult.maxCohesionParameter );
	}
	
	public double[] getScore(int x, int y) {
		int[] neighbCount = new int[ numCultures ];
		int totalNeigh = 0;
		
		totalNeigh += GetInfluenceOfAgents(neighbCount,x,y);
		totalNeigh += GetInfluenceOfVenues(neighbCount,x,y);
		
		if (totalNeigh==0)
			return new double[] { 0, 0 };
		else
			return new double[] { neighbCount[this.cult.identityCode] / (float) totalNeigh, neighbCount[this.cult.identityCode] };
	}
	
	/**
	 * 
	 * Calculates the influence of neighbours, bounces the count in the ref'd array, and returns the total number of neighbours 
	 * 
	 */
	public int GetInfluenceOfAgents(int[] n, int x, int y) {
		int tN = 0;

		int cn = (int) cult.neighbourhoodDist;

		// we are always stepping over the moore neighborhood, so von neumann will be a special check
		for (int a = x - cn; a < x + cn + 1; a++) {
			for (int b = y - cn; b < y + cn + 1; b++) {
				if ( a >= 0 && a < occupiedBy.length && b >= 0 && b < occupiedBy[0].length && (a!=x||b!=y) && occupiedBy[a][b]!=null) {
					
					if (!vonNeumann) {
						tN++;
						n[occupiedBy[a][b].cult.identityCode]++;
					} else {
						// here at last, we'll check the distance
						if (Math.sqrt((x-a)*(x-a) + (y-b)*(y-b)) <= cn) {
							tN++;
							n[occupiedBy[a][b].cult.identityCode]++;
						}
					}
				}
			}
		}		
		return tN;
	}
	
	/**
	 * 
	 * Calculates the influence of venues, same as above in process.
	 * 
	 */
	public int GetInfluenceOfVenues(int[] n, int x, int y) {
		int tN = 0;
		
		for (Agent_Venue v : this.venues) {
			if ( InRange(v, x, y) ) {								
				if ( this.rndOblig > 1 - v.oblig ) {						
					if ( v.cult.identityCode == this.cult.identityCode || this.rndExclu > v.exclu ) {
						
						for (Agent ha : v.visitors) {									
							n[ha.cult.identityCode]++;
						}
						tN += v.visitors.size();													// and then all of them here.
					} 
				}
			}
		}
		return tN;
	}
	
	public int GetInfluenceOfVenuesVisited(int[] n) {
		int tN = 0;
		
		for (Agent_Venue v : this.venuesVisited) {
			for (Agent ha : v.visitors) {	
				n[ha.cult.identityCode]++;								
			}
			tN += v.visitors.size();										// and then all of them here.
		}
		return tN;
	}
}
