package web_server.sim.schelling_180206;

import java.util.ArrayList;
import java.util.UUID;

import processing.core.PVector;
import processing.data.JSONArray;
import processing.data.JSONObject;

public class Agent_Venue extends Agent {
	ArrayList<Agent> visitors = new ArrayList<Agent>();
	Agent[][] occupiedBy;
	float oblig;
	float exclu;
	int travelDist;
	boolean vonNeumann;
	int maxTime;
	int cultureIndex = -1;
	
	int[] store_numVisitors;
	float[] store_fracLocals;
	
	public Agent_Venue (int si, int codeMax, Culture c, Agent[][] occupiedBy, java.util.Random rnd, PVector pv, boolean vn, int maxTime) {
		this.uid = UUID.randomUUID();
		type = "venue";
		if (rnd!=null) {
			pos = new int[] { (int) (rnd.nextDouble() * si), (int) (rnd.nextDouble() * si) };			
		} else {
			pos = new int[] { (int) (pv.x), (int) (pv.y) };
		}

		cult = c;
		cultureIndex = cult.identityCode;
		travelDist = c.venueTravelDist;
		this.maxTime = maxTime;
		this.oblig = c.obligatoriness;
		this.exclu = c.exclusivity;
		this.occupiedBy = occupiedBy;
		this.occupiedBy[pos[0]][pos[1]] = this;
		this.vonNeumann = vn;
		
		store_numVisitors = new int[maxTime];
		store_fracLocals = new float[maxTime];
	}
	
	public Agent_Venue(JSONObject vaa) {
		uid = UUID.fromString(vaa.getString("id"));
		type = "venue";
		pos = new int[] { Integer.parseInt(vaa.getString("pos").split(" ")[0]), Integer.parseInt(vaa.getString("pos").split(" ")[1]) };
		
		store_numVisitors = vaa.getJSONArray("numvis").getIntArray();
		store_fracLocals = vaa.getJSONArray("fracloc").getFloatArray();
		cultureIndex = vaa.getInt("Culture");
	}

	/*
	 * count how many agents up to distance d are of the same culture as this venue
	 */
	public float getFractionLocalAgents(int d) {
		int c = 0;
		int t = 0;

		// we are always stepping over the moore neighborhood, so von neumann will be a special check
		for (int a = pos[0] - d; a < pos[0] + d + 1; a++) {
			for (int b = pos[1]-d; b < pos[1] + d + 1; b++) {
				if ( a >= 0 && a < occupiedBy.length && b >= 0 && b < occupiedBy[0].length && (a!=pos[0]||b!=pos[1]) && occupiedBy[a][b]!=null) {
					if (!vonNeumann) {
						t++;
						if (occupiedBy[a][b].cult.identityCode == this.cult.identityCode) {
							c++;
						}
					} else {
						// here at last, we'll check the distance
						if (Math.sqrt((pos[0]-a)*(pos[0]-a) + (pos[1]-b)*(pos[1]-b)) <= d) {
							t++;
							if (occupiedBy[a][b].cult.identityCode == this.cult.identityCode) {
								c++;
							}
						}
					}
				}
			}
		}
		
		if (t != 0)
			return c / (float) t;
		else
			return 0;
	}
	
	public void Update ( int t ) {
		visitors = new ArrayList<Agent>();
	}
	
	public void StoreStep(int t) {
		store_numVisitors[t] = visitors.size();
		store_fracLocals[t] = getFractionLocalAgents(this.travelDist);
	}

	public JSONObject getJSON() {
		JSONObject jo = new JSONObject();
		jo.setString("id", this.uid.toString());
		jo.setString("pos", this.pos[0] + " " + this.pos[1]);
		jo.setInt("Culture", this.cult.identityCode);
		JSONArray jnumvis = new JSONArray();
		JSONArray jfracloc = new JSONArray();
		for (int i = 0; i < maxTime; i++) {
			jnumvis.append(store_numVisitors[i]);
			jfracloc.append(store_fracLocals[i]);
		}
		jo.setJSONArray("numvis", jnumvis);
		jo.setJSONArray("fracloc", jfracloc);
		return jo;
	}
}
