package web_server.sim.schelling_180206;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import javax.swing.JOptionPane;
import controlP5.*;
import org.gicentre.utils.stat.*;
import processing.core.PApplet;
import processing.core.PFont;
import processing.core.PGraphics;
import processing.core.PVector;
import processing.data.JSONArray;
import processing.data.JSONObject;
import processing.event.MouseEvent;
import web_server.sim.schelling_180206.Sim_Context;

public class Schelling_Interface extends PApplet implements Runnable {	
	PGraphics gWorld;
	PVector gWPos = new PVector(0,270);
	
	PGraphics gAnalysis;
	PVector gAPos = new PVector(510,270);
	String[] AnalysisTypes = new String[] { "World Space", "Parameter Space", "Chart", "Parameter Chart" };
	int aT = 2;																// analysis type overall and vis layer.
	int[] vT = new int[] { 0,0,0,0 };											// for toggling through the vis'
	
	String[] aParams = new String[] {"", ""};								// when we map something in parameter space, these are the dimensions	
	String[][] vNames = new String[][] {
						{ "Frequency of Agent 1", "Frequency of Agent 2", "Moves This Turn","Morphology","Marching Squares","Clusters" },
						
						{ "Fraction of Moore 1 Neighbours, Same Culture [Averaged]", 
						  "Num. Unhappy Agents",
						  "Fraction Same Culture in Travel Distance of Venue",
						  "Num. Clusters", 
						  "Perim. Area Ratio of Clusters", 
						  "Morphological Stability"				},
						
						{ "Diversity of Neighborhood", "Num. Unhappy Agents","Num. Moved Agents", "Fraction Same Culture in Travel Dist.", 
							"Morphological Stability","Num. Clusters", "Perimeter-Area Ratio of Clusters" },
						
						{ "Fraction of Moore 1 Neighbours, Same Culture [Averaged] in terms of _ ", 
						  "Fraction of Same Culture in Travel Distance of Venue",
						  "Num. Clusters",
						  "Perim. Area Ratio of Clusters",
						  "Morphological Stability"				}};
	PGraphics gForm;
	PVector gFPos = new PVector(762,10);
	String[] FormTypes = new String[] { "Venues" };
	
	PGraphics gDistri;
	PVector gDPos = new PVector(510,10);
	
	boolean venDrag = false;
	PVector activeVenue = null;
	boolean expiredSolution = false;
	
	boolean redrawViews = true;
	
	boolean loadView = false;
	int loadStep = 0;
	ArrayList<String> loadBits = new ArrayList<String>();
	String[] fileList = new String[0];
	String[] pNames = new String[] { "Threshold", "MaxThreshold", "Density", "Obligatoriness", "Exclusivity", "NumVenues", "NeighDist", "VenDist" };

	boolean running = false;			// for progress bar
	int currentRun = 0;
	
	ArrayList<XYChart> lineChart;
	ControlP5 cp5;
	
	float[][] mpa = null;
	PFont mf;
	
	SingleRun[] sr = new SingleRun[1];
	SingleRun[] srNew;
	int ai = 0;
	int maxTime = 20;
	int time = 1;

	ArrayList<Controller> cp5Main = new ArrayList<Controller>();
	ArrayList<Controller> cp5Save = new ArrayList<Controller>();
	ArrayList<Controller> cp5Load = new ArrayList<Controller>();
	
	int[] colours = new int[ ] { color(255,0,0), color(0,0,255), 
								 color(255,255,0), color(0,255,0), 
								 color(255,215,0), color(255,192,203) };
	int[] coloursA = new int[ ] { color(255,128,128), color(128,128,255), 
								  color(255,255,128), color(128,255,128), 
								  color(255,215,128), color(255,192,203) };

	String baseFolder = "20180522//";
	String activeFolder = "Temp";
	
	Sim_ReadWrite srw = new Sim_ReadWrite( this, baseFolder );

	public static void main(String[] args) {
		PApplet.main( "web_server.sim.schelling_180206.Schelling_Interface" );
	}
	
	public void settings() {
        size(1300,770);
	}
	
	public void setup() {
		hint(DISABLE_ASYNC_SAVEFRAME);
		background(255);
		
		gWorld = createGraphics(501,501);
		gAnalysis = createGraphics(501,501, JAVA2D);

		gForm = createGraphics(251,251);
		gDistri = createGraphics(251,251);
		textSize(8);
		mf = createFont("Arial",10);
		addUI();
		
		srw.CreateOrClearTempFolder();						// clears anything from temp during the last run of the program
	}
	
	public void draw() {
		background(255);
		textSize(9);

		if (!running) {
			if (sr[ai]!=null) {			
				if (redrawViews) {
					redrawViews = false;									// so that it isn't being redrawn every frame
					println(ai + " " + time + " " + sr.length);
					
					DrawWorldAtT(gWorld, sr[ai], time);
					DrawVenuesAtT(gForm, sr[ai], time);
					DrawDistributionAtT(gDistri, sr[ai], 0);
					DrawAnalysis();
				}
			}

			image(gWorld,gWPos.x,gWPos.y);
			image(gForm,gFPos.x, gFPos.y);
			image(gDistri,gDPos.x, gDPos.y);
			image(gAnalysis,gAPos.x,gAPos.y);

			// draw analysis mouse-over
			if (mouseOnAnalysis() && aT==1) {
				fill(255);
				rect(mouseX+25, mouseY+25, 90, 14);
				fill(0);
				text( getDataFromMouseOverAnalysis((int)(mouseX-gAPos.x),(int)(gAnalysis.height-(mouseY-gAPos.y))), mouseX+35, mouseY+35);
			} else if (mouseOnAnalysis() && aT==2) {
				PVector pv = lineChart.get(ai).getScreenToData(new PVector(mouseX-gAPos.x, /*gAnalysis.height-*/(mouseY-gAPos.y)));
				if (pv!=null&& Math.abs(pv.x-(int)pv.x) < 0.3f ) {
					float f = lineChart.get(ai).getYData()[(int)pv.x];
					String s = "" + f;
					text( s, gAPos.x+10, gAPos.y+10);
					PVector pvv = lineChart.get(ai).getDataToScreen(new PVector((int)(pv.x),f));
					ellipse(gAPos.x+pvv.x,gAPos.y+pvv.y,5,5);
				}
			}
			
			if (mouseOnWorld()) {
				fill(255);
				rect(mouseX+25,mouseY+25, 100,100 );
				fill(0);
				text ( getDataFromMouseOverWorld ( (int) (mouseX-gWPos.x), (int) (gWorld.height - (mouseY-gWPos.y)) ), mouseX+35, mouseY+35 );
			}

			if (expiredSolution) {
				noStroke();
				fill(255,150);
				rect(gWPos.x,gWPos.y,gWorld.width, gWorld.height);
				rect(gAPos.x,gAPos.y,gAnalysis.width, gAnalysis.height);
			}

			if (loadView) {
				for (int i = 0; i < 3; i++) {
					noFill();
					stroke(0);
					rect (width-250, 50 + i * 150, 245, 145);	

					if (i + loadStep < fileList.length && i + loadStep > -1) {	
						fill(0);
						String[] st = fileList[i+loadStep].split(",");
						text ("Name: " + st[0] + "\nNum. Runs: " + st[1] + "\nAuthor: " + st[2] + "\nDescription: " + st[3],  width-250, 50 + i * 150, 245, 145 );
						noFill();
					}
				}
			}
			cp5.draw();
		} else {
			if (srNew!=null) {
				fill(0);
				text ("running " + (currentRun+1) + " / " + srNew.length, width/2, height/2);
			}
		}
		
		if (!loadView)
			HideLoadingUI();
	}
	
	public void DrawAnalysis() {
		if (aT==0) {
			if (vT[aT]==0) {
				DrawProbabilityAtT(gAnalysis,"h 0", sr, time);
			} else if (vT[aT]==1) {
				DrawProbabilityAtT(gAnalysis,"h 1", sr, time);
			} else if (vT[aT]==2) {
				DrawMovesAtT(gAnalysis, sr, time);
			} 
		} else if (aT==1) {
			DrawParameterSpace(gAnalysis, sr, time);	
		} else if (aT==2) {
			addCharts(gAnalysis);
		} else if (aT==3) {
			addParaCharts(gAnalysis);
		}
	}
	
	public XYChart getChartFromParam(XYChart lc, int aC) {
		String xAxis = "";
		PVector xR = new PVector(0,0);
		
		ArrayList<PVector> data = new ArrayList<PVector>();
		
		if (!aParams[0].equals("")) {
			xAxis = aParams[0];
			Range xRange = (Range) cp5.getController( aParams[0].substring(0,aParams[0].length()-1) + "r" );
			xR = new PVector ( xRange.getLowValue(), xRange.getHighValue() );
		}
		int numStepsX = xAxis.equals("") ? 1 : (int) cp5.getController("Num_Steps").getValue();
		
		for (int i = 0; i < numStepsX; i++) {
			float xVal = 0;	
			xVal = xR.x + (xR.y-xR.x) * (i/(float) (numStepsX-1));
			int num = 0;
			float tot = 0;

			for (SingleRun srr : sr) {
				if ( (xAxis.equals("") || Math.abs( srr.jsP.getJSONObject(xAxis.charAt(0)+"_settings").getFloat( xAxis.split("_")[1]) - xVal) < 0.001) ) {
					num++;
					
					if (aC==0)
						tot += srr.DiversityPerTime[time];
					else if (aC==1)		// same culture in travel dist
						tot += srr.VenFractionPerTime[time];
					else if (aC==2)		// num clusters
						tot += srr.numClusters[time];
					else if (aC==3)		// perim area ratio
						tot += srr.perimArea[time];
					else if (aC==4)		// morphological stability
						tot += srr.mStability[time];
				}
			}
			tot = tot / (float) num;
			data.add(new PVector (i, tot));
		}
		lc.setData( data );
		
		lc.setMinY(0);
		
		lc.setMinX(0);
		lc.setMaxX(numStepsX);
		if (aC==0)		{ lc.setYAxisLabel("Homogeneity");	lc.setMaxY(1); }
		else if (aC==1)	{ lc.setYAxisLabel("Fraction");		lc.setMaxY(1); }
		else if (aC==2)	{ lc.setYAxisLabel("Clusters");		lc.setMaxY(20); }
		else if (aC==3)	{ lc.setYAxisLabel("Ratio");		lc.setMaxY(5); }
		else if (aC==4)	{ lc.setYAxisLabel("Stability");	lc.setMaxY(1); }
		
		lc.setXAxisLabel(xAxis);		
		
		return lc;
	}
	
	public void addParaCharts (PGraphics pg) {		
		pg.beginDraw();
		pg.textFont(mf, 9);
		pg.background(255);
		pg.noFill();
		pg.rect(0, 0, pg.width-1, pg.height-2);
		
		XYChart lc = new XYChart(this);
		lc.setGraphics(pg);

		lc = getChartFromParam(lc, vT[3]);
		
		lc.setPointColour(color(180,50,50));
		lc.setLineColour(color(0,0,0));
		lc.setPointSize(3);
		lc.setLineWidth(1);
		lc.showXAxis(true); 
		lc.showYAxis(true); 
		lc.draw(15,200,450,250);
		pg.endDraw();
	}
	
	public void addCharts(PGraphics pg) {
		lineChart.clear();
		
		pg.beginDraw();
		pg.textFont(mf, 9);
		pg.background(255);
		pg.noFill();
		pg.rect(0, 0, pg.width-1, pg.height-2);
		
		int k = 0;
		// draw all of the charts
		for (SingleRun srr : sr)
			addOneChart(srr, pg, k++);
		
		// draw the currently active chart more intensely
		XYChart lc = lineChart.get(ai);
  	  	lc.setLineColour(color(0,0,0,255));
  	  	lc.setLineWidth(1f);
  	  	for (XYChart lcc : lineChart) {
  	  		lcc.draw(15,200,500,250);		
  	  	}
  	  	
  	  	// indicate the current time
		if (lineChart.get(0)!=null) {
			PVector lpv = lineChart.get(0).getDataToScreen(new PVector(time, 0));
			gAnalysis.line(lpv.x,lpv.y, lpv.x, lpv.y-220);
		}
		pg.endDraw();
	}
	
	public void addOneChart (SingleRun srz, PGraphics pg, int k) {
		XYChart lc = new XYChart(this);
		lc.setGraphics(pg);
		int activeChart = vT[2];
		
		if (activeChart==0) {
			lc.setData(srz.getDiversityChart());
			lc.setMinY(0);
			lc.setMaxY(1);
			lc.setMinX(0);
			lc.setMaxX(maxTime+1);
			if (k == 0) {
				lc.setYAxisLabel("Diversity");
				lc.setXAxisLabel("Time");
			} else {
				lc.setYAxisLabel(" ");
				lc.setXAxisLabel(" ");
			}
		} else if (activeChart==1) {
			lc.setData(srz.getUnhappinessChart());
			lc.setMinY(0);
			lc.setMaxY(2500);
			lc.setMinX(0);
			lc.setMaxX(maxTime+1);
			if (k == 0) {
				lc.setYAxisLabel("Num. Unhappy Agents");
				lc.setXAxisLabel("Time");
			} else {
				lc.setYAxisLabel(" ");
				lc.setXAxisLabel(" ");
			}
		} else if (activeChart==2) {
			lc.setData(srz.getMovedChart());
			lc.setMinY(0);
			lc.setMaxY(2500);
			lc.setMinX(0);
			lc.setMaxX(maxTime+1);
			
			if (k == 0) {
				lc.setYAxisLabel("Num. Agents that Moved");
				lc.setXAxisLabel("Time");
			} else {
				lc.setYAxisLabel(" "); 
				lc.setXAxisLabel(" "); 	
			}
		} else if (activeChart==3) {
			lc.setData(srz.getVenueNeighborhoods());
			lc.setMinY(0);
			lc.setMaxY(1);
			lc.setMinX(0);
			lc.setMaxX(maxTime+1);
			
			if (k == 0) {
				lc.setYAxisLabel("Fraction of Venue Neighborhood");
				lc.setXAxisLabel("Time");
			} else {
				lc.setYAxisLabel(" ");
				lc.setXAxisLabel(" ");
			}
		} else if (activeChart==4) {
			lc.setData( srz.getStability() );
			lc.setMinY(0);
			lc.setMaxY(1);
			lc.setMinX(0);
			lc.setMaxX(maxTime+1);
			
			if (k == 0) {
				lc.setYAxisLabel("Morphological Stability");
				lc.setXAxisLabel("Time");
			} else {
				lc.setYAxisLabel(" ");
				lc.setXAxisLabel(" ");
			}
		} else if (activeChart==5) {
			lc.setData( srz.getNumClusters() );
			lc.setMinY(0);
			lc.setMaxY(50);
			lc.setMinX(0);
			lc.setMaxX(maxTime+1);
			
			if (k == 0) {
				lc.setYAxisLabel("Num. Clusters");
				lc.setXAxisLabel("Time");
			} else {
				lc.setYAxisLabel(" ");
				lc.setXAxisLabel(" ");
			}
		} else if (activeChart==6) {
			lc.setData( srz.getPerimArea() );
			lc.setMinY(0);
			lc.setMaxY(5);
			lc.setMinX(0);
			lc.setMaxX(maxTime+1);
			
			if (k == 0) {
				lc.setYAxisLabel("Perimeter-Area Ratio of Clusters");
				lc.setXAxisLabel("Time");
			} else {
				lc.setYAxisLabel(" ");
				lc.setXAxisLabel(" ");
			}
		}
		
		lc.setPointColour(color(180,50,50,100));
		lc.setLineColour(color(0,0,0,100));
		lc.setPointSize(2);
		lc.setLineWidth(.5f);		

		lc.showXAxis(true); 
		lc.showYAxis(true); 
		
		lineChart.add(lc);
	}
	
	// load up this run from file, since we're trying to ask something about it more specfically
	public SingleRun ThickenUp (JSONObject job) {
		return new SingleRun(job.getJSONObject("Settings"), job.getJSONObject("Results"), job.getJSONObject("Analysis"), false );
	}

	
	public void setIteration ( int last, int cur ) {
		// unload / lighten up the current iteration, thicken the new one
		if (cur!=last) {
			if (last >= 0)
				sr[last].LightenUp();
			
			sr[cur] = ThickenUp(loadJSONObject(baseFolder + activeFolder + "//" + cur + ".txt"  ));

			cp5.getController("it").setCaptionLabel(ai+" / "+(sr.length));		
			println("here");
			
			redrawViews = true;
			UpdateUILabels();
		}
	}

	public void setTime(int c) {
		c = max(0, min(maxTime-1,c));
		time = c;
		
		if (cp5.getController("t")!=null) {
			cp5.getController("t").setCaptionLabel( time + " / " + (maxTime-1) );
		}
		redrawViews = true;
	}
	
	public boolean mouseOnVenues() {
		return mouseX > gFPos.x && mouseX < gFPos.x + gForm.width && mouseY > gFPos.y && mouseY < gFPos.y + gForm.height;
	}
	
	public boolean mouseOnAnalysis() {
		return mouseX > gAPos.x && mouseX < gAPos.x + gAnalysis.width && mouseY > gAPos.y && mouseY < gAPos.y + gAnalysis.height;
	}
	
	public boolean mouseOnWorld() {
		return mouseX > gWPos.x && mouseX < gWPos.x + gWorld.width && mouseY > gWPos.y && mouseY < gWPos.y + gWorld.height;
	}
	
	public void mousePressed() {
		if ( mouseOnVenues() ) {
			float cs = gForm.width / (float) sr[ai].cells[0].length;
			PVector mP = new PVector ( (mouseX - gFPos.x)/cs  , (mouseY - gFPos.y)/cs  );

			for (PVector pv : sr[ai].VenueLocations) {
				if (pv.dist(mP) < 2 ) {
					venDrag = true;
					activeVenue = pv;
					break;
				}
			}
			redrawViews = true;
		}
		
	}
	
	public void mouseReleased() {
		if ( mouseOnVenues() && venDrag && activeVenue!=null ) {
			float cs = gForm.width / (float) sr[ai].cells[0].length;
			
			activeVenue.x = (int) (mouseX - gFPos.x)/(int)cs;
			activeVenue.y = (int) (mouseY - gFPos.y)/(int)cs;
			expiredSolution = true;
			redrawViews = true;
		}
		venDrag = false;
		activeVenue = null;		
		
	}
	
	public void mouseClicked(MouseEvent evt) {
		if (evt.getCount() == 2)
			doubleClicked();
		
		if (mouseOnAnalysis()) {
			SetActiveFromClick((int)(mouseX-gAPos.x),gAnalysis.height-(int)(mouseY-gAPos.y));
		}		
	}
	
	public void SetActiveFromClick (int x, int y) {
		String xAxis = aParams[0];
		String yAxis = aParams[1];
		PVector xR = new PVector(0,0);
		PVector yR = new PVector(0,0);
		
		int numSteps = (int) cp5.getController("Num_Steps").getValue();
		int siX = (gAnalysis.width-1) / (xAxis.equals("") ? 1 : numSteps);
		int siY = (gAnalysis.width-1) / (yAxis.equals("") ? 1 : numSteps);
		x = x/siX;
		y = y/siY;
		
		if (!xAxis.equals("")) {
			Range xRange = (Range) cp5.getController( aParams[0].substring(0,aParams[0].length()-1) + "r" );
			xR = new PVector ( xRange.getLowValue(), xRange.getHighValue() );
		}
		
		if (!yAxis.equals("")) {
			Range yRange = (Range) cp5.getController( aParams[1].substring(0,aParams[1].length()-1) + "r" );
			yR = new PVector ( yRange.getLowValue(), yRange.getHighValue() );
		}
		
		float xVal = xR.x + (xR.y-xR.x) * x/(numSteps-1);
		float yVal = yR.x + (yR.y-yR.x) * y/(numSteps-1);
				
		int z = 0;
		for (SingleRun srr : sr) {
			if ( (xAxis.equals("") || Math.abs( srr.jsP.getJSONObject(xAxis.charAt(0)+"_settings").getFloat( xAxis.split("_")[1]) - xVal) < 0.001) 
					&& (yAxis.equals("") || Math.abs( srr.jsP.getJSONObject(yAxis.charAt(0)+"_settings").getFloat( yAxis.split("_")[1]) - yVal) < 0.001 )) {
				int aai = ai;
				ai = z;
				setIteration(aai,ai);
				break;
			}
			z++;
		}
	}
	
	public String getDataFromMouseOverWorld (int x, int y) {
		
		return "";
	}
	
	public String getDataFromMouseOverAnalysis( int x, int y ) {
		String xAxis = aParams[0];
		String yAxis = aParams[1];
		PVector xR = new PVector(0,0);
		PVector yR = new PVector(0,0);
		
		int numSteps = (int) cp5.getController("Num_Steps").getValue();
		int siX = (gAnalysis.width-1) / (xAxis.equals("") ? 1 : numSteps);
		int siY = (gAnalysis.width-1) / (yAxis.equals("") ? 1 : numSteps);
		x = x/siX;
		y = y/siY;
		
		if (!xAxis.equals("")) {
			Range xRange = (Range) cp5.getController( aParams[0].substring(0,aParams[0].length()-1) + "r" );
			xR = new PVector ( xRange.getLowValue(), xRange.getHighValue() );
		}
		
		if (!yAxis.equals("")) {
			Range yRange = (Range) cp5.getController( aParams[1].substring(0,aParams[1].length()-1) + "r" );
			yR = new PVector ( yRange.getLowValue(), yRange.getHighValue() );
		}
		
		float xVal = xR.x + (xR.y-xR.x) * x/(numSteps-1);
		float yVal = yR.x + (yR.y-yR.x) * y/(numSteps-1);

		DecimalFormat formatter = new DecimalFormat("#0.00");
		
		if (mpa!=null && x < mpa.length && y < mpa[x].length)
			return formatter.format(xVal).toString() + "," + formatter.format(yVal).toString() + " : " + formatter.format(mpa[x][y]);
		else
			return "";
	}
	
	public void doubleClicked() {
		float cs = gForm.width / (float) sr[ai].cells[0].length;
		
		if ( mouseOnVenues() ) {
			PVector mP = new PVector ( (mouseX - gFPos.x)/cs, (mouseY - gFPos.y)/cs );

			for (PVector pv : sr[ai].VenueLocations) {
				if (pv.dist(mP) < 2 ) {
					String s = JOptionPane.showInputDialog("Enter coordinates in the form of X,Y. Origin is 0,0 at the top-left, max value is WorldSize-1");
					
					pv.x = Integer.parseInt(s.split(",")[0]);
					pv.y = Integer.parseInt(s.split(",")[1]);
					break;
				}
			}
		}
	}
	
	public void Run_Sim(int aa) {
		running = true;
		(new Thread( this )).start();
	}

	public void run() {
		activeFolder = "Temp";					// if run has been hit, we're back to doing something new rather than browsing from file...
		srw.WipeTemp();
				
		lineChart = new ArrayList<XYChart>();
		maxTime = (int) cp5.getController("Time_Steps").getValue();
		
		int numRanges = getNumRanges();											// the number of active variables
		int iter = (int) cp5.getController("Num_Iterations").getValue();		// number of iterations with the same parameters
		int numSteps = (int) cp5.getController("Num_Steps").getValue();			// moving from a min to a max, we'll make this many steps
		
		if (srNew==null)
			srNew = new SingleRun[ iter * (int) pow(numSteps, numRanges) ];
		
		// the length of sr should be iterations x num_steps^num_ranges
		for (int z = 0; z < srNew.length; z++) {
			currentRun = z;															// for progress bar
			Sim_Context sc = new Sim_Context( );			
			JSONObject jsSettings = GenerateSettings(numSteps,numRanges,iter,z);		

			JSONObject jsAllOutcomes = sc.run ( jsSettings ); 				// will take in a single JSON, 
			
			JSONObject jsResults = jsAllOutcomes.getJSONObject("Results");
			JSONObject jsAnalysis = jsAllOutcomes.getJSONObject("Analysis");
			
			SingleRun srr = new SingleRun(jsSettings, jsResults, jsAnalysis, z!=0 );		// is it OK for this to be the "light version"?
			
			srNew[z] = srr;
			jsAllOutcomes.setJSONObject("Settings", jsSettings);
			srw.TempSaveText(z, jsAllOutcomes);	// instead of the singlerun
			
			// it's already light
			//if (z != 0) {
			//	srNew[z].LightenUp();
			//}
			sc = null;
		}
		sr = srNew;
		srNew = null;
	
		expiredSolution = false;
		setTime(0);
		ai = 0;
		setIteration(-1, ai);

		running = false;
	}
	
	
	public void World_Size(float ws) {
		cp5.getController("World_Size").changeValue( floor(ws) );
	}
	public void Num_Cultures(float ws) {
		cp5.getController("Num_Cultures").changeValue( floor(ws) );
		UpdateControlsVis();
	}
	public void Time_Steps(float ws) {
		cp5.getController("Time_Steps").changeValue( floor(ws) );
	}
	
	public void DrawParameterSpace (PGraphics pg, SingleRun[] sRuns, int t) {
		// find out what the 1 or 2 active parameters are, chart them on x and y, sizing the rectangles according to the number of each...
		// i can use their parameter ranges to get the values that 
		PVector highlightCurrent = new PVector(-1,-1);
		
		String xAxis = "";
		PVector xR = new PVector(0,0);
		String yAxis = "";
		PVector yR = new PVector(0,0);
		
		if (!aParams[0].equals("")) {
			xAxis = aParams[0];
			Range xRange = (Range) cp5.getController( aParams[0].substring(0,aParams[0].length()-1) + "r" );
			xR = new PVector ( xRange.getLowValue(), xRange.getHighValue() );
		}
		
		if (!aParams[1].equals("")) {
			yAxis = aParams[1];
			Range yRange = (Range) cp5.getController( aParams[1].substring(0,aParams[1].length()-1) + "r" );
			yR = new PVector ( yRange.getLowValue(), yRange.getHighValue() );
		}
		int numStepsX = xAxis.equals("") ? 1 : (int) cp5.getController("Num_Steps").getValue();
		int numStepsY = yAxis.equals("") ? 1 : (int) cp5.getController("Num_Steps").getValue();
		float siX = (pg.width-1) / (float) numStepsX;
		float siY = (pg.width-1) / (float) numStepsY;
		
		mpa = new float[numStepsX][numStepsY];
		
		pg.beginDraw();
		pg.background(255);
		
		for (int i = 0; i < numStepsX; i++) {
			float xVal = 0;	
			xVal = xR.x + (xR.y-xR.x) * (i/(float) (numStepsX-1));
			
			for (int j = 0; j < numStepsY; j++) {
				float yVal = 0;
				yVal = yR.x + (yR.y-yR.x) * (  (numStepsY-1-j)  /   (float)(numStepsY-1));
				
				int num = 0;
				float tot=0;
				
				int z = 0;
				for (SingleRun sr : sRuns) {
					if ( (xAxis.equals("") || Math.abs( sr.jsP.getJSONObject(xAxis.charAt(0)+"_settings").getFloat( xAxis.split("_")[1]) - xVal) < 0.001) 
							&& (yAxis.equals("") || Math.abs( sr.jsP.getJSONObject(yAxis.charAt(0)+"_settings").getFloat( yAxis.split("_")[1]) - yVal) < 0.001 )) {
						num++;
						
						if (vT[aT]==0)
							tot += sr.DiversityPerTime[t];
						else if (vT[aT]==1)
							tot += sr.UnsettledPerTime[t];
						else if (vT[aT]==2) 
							tot += sr.VenFractionPerTime[t];
						else if (vT[aT]==3)
							tot += sr.numClusters[t];
						else if (vT[aT]==4)
							tot += sr.perimArea[t];
						else if (vT[aT]==5)
							tot += sr.mStability[t];
						
						if (ai == z) {
							highlightCurrent = new PVector(i, j);
						}
					}
					z++;
				}
				tot = tot / (float) num;
				mpa[i][numStepsY-1-j] = tot;
				if (vT[aT]==0 || vT[aT]==2 || vT[aT]==5)
					pg.fill(255*tot);
				else if (vT[aT]==1)
					pg.fill(255 * tot/1000f);
				else if (vT[aT]==3)
					pg.fill(255 * tot/25f);
				else if (vT[aT]==4)
					pg.fill(255 * tot / 2f);
				
				pg.rect(i*siX,j*siY,siX,siY);
			}
		}
		pg.noFill();
		pg.strokeWeight(2);
		pg.stroke(255,0,0);
		pg.rect(highlightCurrent.x*siX, highlightCurrent.y*siY, siX, siY);
		pg.stroke(0);
		pg.strokeWeight(1);
		// highlight the current iteration
		pg.endDraw();
	}

	
	
	public void DrawWorldAtT (PGraphics pg, SingleRun sr, int t) {
		float cs = (pg.width-1) / (float) sr.cells[0].length;
		
		pg.beginDraw();
		pg.background(255);
		pg.stroke(0);
		pg.strokeWeight(0.5f);
		
		for (int i = 0; i < sr.cells[0].length; i++) {
			for (int j = 0; j < sr.cells[0][0].length; j++) {
				pg.stroke(0);
				pg.strokeWeight(0.5f);
				
				if ( sr.cells[t][i][j].equals("0") ) {
					pg.fill(255);
					pg.rect ( i*cs,j*cs,cs,cs );
					
				} else if ( sr.cells[t][i][j].contains("h") ) {
					// draw an agent
					pg.fill ( colours[ Integer.parseInt( sr.cells[t][i][j].split(" ")[1] ) ], 255*0.7f);
					pg.rect ( i*cs,j*cs,cs,cs );
					
				} else if ( sr.cells[t][i][j].contains("v")  ) {
					// draw a venue
					pg.noFill();
					pg.stroke( colours[ Integer.parseInt( sr.cells[t][i][j].split(" ")[1] ) ]);
					pg.strokeWeight(2);
					pg.line(i*cs,j*cs,(i+1)*cs,(j+1)*cs);
					pg.line((i+1)*cs,j*cs,(i)*cs,(j+1)*cs);
				}
				
			}
		}
		pg.endDraw();
	}
	
	public void DrawVenuesAtT(PGraphics pg, SingleRun sr, int t) {
		float cs = (pg.width-1) / (float) sr.cells[0].length;
		
		pg.beginDraw();
		pg.background(255);
		pg.stroke(0);
		pg.strokeWeight(0.5f);
		pg.noFill();
		
		for (int i = 0; i < sr.cells[0].length; i++) {
			for (int j = 0; j < sr.cells[0][0].length; j++) {
				pg.rect ( i*cs,j*cs,cs,cs );
			}
		}
		
		for (PVector pv : sr.VenueLocations) {
			pg.fill (colours[ (int) pv.z]);
			pg.rect ( pv.x*cs,pv.y*cs,cs,cs );
		}
		pg.endDraw();
	}
	
	public void DrawDistributionAtT( PGraphics pg, SingleRun sr, int t ) {
		float cs = (pg.width-1) / (float) sr.cells[0].length;
		
		pg.beginDraw();
		pg.background(255);
		pg.stroke(0);
		pg.strokeWeight(0.5f);
		pg.noFill();
		
		for (int i = 0; i < sr.cells[0].length; i++) {
			for (int j = 0; j < sr.cells[0][0].length; j++) {
				pg.stroke(0);
				pg.strokeWeight(0.5f);
				
				if ( sr.cells[t][i][j].equals("0") ) {
					pg.fill(255);
					pg.rect ( i*cs,j*cs,cs,cs );
					
				} else if ( sr.cells[t][i][j].contains("h") ) {
					// draw an agent
					pg.fill ( colours[ Integer.parseInt( sr.cells[t][i][j].split(" ")[1] ) ], 255*0.7f);
					pg.rect ( i*cs,j*cs,cs,cs );
					
				}
			}
		}
		pg.endDraw();
	}
	
	public void DrawProbabilityAtT(PGraphics pg, String target, SingleRun[] sRuns, int t) {

	}
	
	public void DrawMovesAtT(PGraphics pg, SingleRun[] sRuns, int t) {
		float cs = pg.width / (float) sr[ai].cells[0].length;
		
		pg.beginDraw();
		pg.background(255);
		pg.stroke(0);
		pg.strokeWeight(0.5f);
		pg.noFill();
		
		
		for (int i = 0; i < sr[ai].cells[t].length; i++) {
			for (int j = 0; j < sr[ai].cells[t][0].length; j++) {
				pg.rect ( i*cs,j*cs,cs,cs );
			}
		}
		
		pg.strokeWeight(1);
		
		for (Agent_Human ha : sRuns[ai].ha) {
			pg.stroke( colours[ha.cultureIndex],100 );
			if ( ha.store_moved[t] ) {
				pg.line(ha.store_pos[t][0]*cs + cs/2, ha.store_pos[t][1]*cs + cs/2, ha.store_pos[t-1][0]*cs + cs/2, ha.store_pos[t-1][1]*cs + cs/2);
				pg.fill(colours[ha.cultureIndex],100);
				pg.ellipse(ha.store_pos[t][0]*cs + cs/2, ha.store_pos[t][1]*cs + cs/2,cs/4f,cs/4f);
				pg.noFill();
			}
		}
		pg.endDraw();
	}
	
	public void DrawObligation (PGraphics pg, SingleRun[] sRuns, int t) {
		float cs = pg.width / (float) sr[ai].cells[0].length;
		
		pg.beginDraw();
		pg.background(255);
		pg.stroke(0);
		pg.strokeWeight(0.5f);
		pg.fill(0,255,0);
		//pg.noFill();
		
		for (int i = 0; i < sr[ai].cells[t].length; i++) {
			for (int j = 0; j < sr[ai].cells[t][0].length; j++) {
				pg.rect ( i*cs,j*cs,cs,cs );
			}
		}
		
		pg.strokeWeight(1);
		
		for (Agent_Human ha : sRuns[ai].ha) {
			pg.fill( ha.store_rndOblig[t] * 255);
			pg.rect(ha.store_pos[t][0]*cs, ha.store_pos[t][1]*cs,cs,cs);
		}
		pg.endDraw();
	}
	
	public void DrawExclusivity (PGraphics pg, SingleRun[] sRuns, int t) {
		float cs = pg.width / (float) sr[ai].cells[0].length;
		
		pg.beginDraw();
		pg.background(255);
		pg.stroke(0);
		pg.strokeWeight(0.5f);
		pg.fill(0,255,0);
		
		for (int i = 0; i < sr[ai].cells[t].length; i++) {
			for (int j = 0; j < sr[ai].cells[t][0].length; j++) {
				pg.rect ( i*cs,j*cs,cs,cs );
			}
		}
		pg.strokeWeight(1);
		
		for (Agent_Human ha : sRuns[ai].ha) {
			pg.fill( ha.store_rndExclu[t] * 255);
			pg.rect(ha.store_pos[t][0]*cs, ha.store_pos[t][1]*cs,cs,cs);
		}
		pg.endDraw();
	}
	
	
	public void mouseWheel(MouseEvent event) {
		if (loadView) {
			loadStep+=event.getCount();
			loadStep = max(0,loadStep);
			loadStep = min(fileList.length-3,loadStep);
		}
	}
	
	public void PublishRun() {
	
	}
	
	public void LoadFromFile( String da, String num ) {
		activeFolder = da;
		// look at the folder location, count the number of files, and load up based on that
		JSONObject jSet = loadJSONObject(baseFolder + da + "//_gSettings.txt");
		sr = new SingleRun[Integer.parseInt(num)];

		for (int i = 0; i < sr.length; i++) {
			JSONObject jo = loadJSONObject(baseFolder + da + "//" + i + ".txt");
			
			// if i!=0, we avoid looking at the results at all... we only need them for 0 because it's the one that'll be drawn to screen
			sr[i] = new SingleRun( jo.getJSONObject("Settings"), i==0 ? jo.getJSONObject("Results") : null, jo.getJSONObject("Analysis"), i!=0);
		}
		
		setTime(0);
		setIteration(-1, 0);
		
		// global and instance settings 		
		SetUIAfterLoad( jSet, sr[0].jsP ); //runs.getJSONObject(0).getJSONObject("Settings"), jo.getJSONArray("GlobalSettings") );
	}
	
	public void UpdateControlsVis() {
		boolean c = ((Button) cp5.getController("Global_Override")).isOn();

		for (int v = 0; v < cp5.getController("Num_Cultures").getMax(); v++) {
			boolean b = v < cp5.getController("Num_Cultures").getValue() && ( v==0 || !c );	

			for (String s : Culture.CultureControls) {
				boolean bx = v>0 && ((Button)cp5.getController("0_" + s + "_x")).isOn();		// local override
				boolean bb = ((Button)cp5.getController(v + "_" + s +"_b")).isOn();	
				
				cp5.getController(v + "_"+ s +"_a").setVisible(b&&!bx);							// active dimension 
				cp5.getController(v + "_"+ s +"_b").setVisible(b&&!bx);							// slider v. range
				cp5.getController(v + "_"+ s +"_t").setVisible(b&&!bx);							// text
				cp5.getController(v + "_"+ s).setVisible(b && !bb && !bx);						// slider
				cp5.getController(v + "_"+ s +"_r").setVisible(b && !bx && bb);					// range
			}
			cp5.getController("Culture "+ v).setVisible(b);
		}
		UpdateUILabels();
	}
		
	public void SetUIAfterLoad( JSONObject jSet, JSONObject jso ) {
		println(jso);
		maxTime = jSet.getInt("Time_Steps");
		
		cp5.getController("Num_Iterations").setValue( jSet.getInt("Num_Iterations")  );
		cp5.getController("Num_Cultures").setValue( jSet.getFloat("Num_Cultures")  );
		cp5.getController("World_Size").setValue( jSet.getInt("World_Size")  );
		cp5.getController("Time_Steps").setValue( jSet.getInt("Time_Steps")  );
		cp5.getController("Num_Steps").setValue( jSet.getInt("Num_Steps"));
		cp5.getController("Random_Moves").setValue(jSet.getFloat("Random_Moves"));
		
		if ( jSet.getBoolean("VN") ) {
			((Button)cp5.getController("VN")).setOn();
			((Button)cp5.getController("M")).setOff();
		} else {
			((Button)cp5.getController("VN")).setOff();
			((Button)cp5.getController("M")).setOn();
		}
		if (jSet.getBoolean("Venue_Lock")) {
			((Button)cp5.getController("Venue_Lock")).setOn();
		} else {
			((Button)cp5.getController("Venue_Lock")).setOff();
		}
		if (jSet.getBoolean("Distribution_Lock")) {
			((Button)cp5.getController("Distribution_Lock")).setOn();
		} else {
			((Button)cp5.getController("Distribution_Lock")).setOff();
		}
		if (jSet.getBoolean("Global_Override")) {
			((Button)cp5.getController("Global_Override")).setOn();
		} else {
			((Button)cp5.getController("Global_Override")).setOff();
		}
		
		if ( jSet.getString("Movement").equals("Best") ) {
			((Button)cp5.getController("Best")).setOn();
			((Button)cp5.getController("Closest")).setOff();
			((Button)cp5.getController("Rand_Works")).setOff();
		} else if (jSet.getString("Movement").equals("Closest") ) {
			((Button)cp5.getController("Closest")).setOn();
			((Button)cp5.getController("Best")).setOff();
			((Button)cp5.getController("Rand_Works")).setOff();
		} else if (jSet.getString("Movement").equals("Rand") ) {
			((Button)cp5.getController("Rand_Works")).setOn();
			((Button)cp5.getController("Best")).setOff();
			((Button)cp5.getController("Closest")).setOff();
		}
		
		if (jSet.getString("Learning").equals("Random")) {
			((Button)cp5.getController("Random")).setOn();
			((Button)cp5.getController("Inertial")).setOff();
			((Button)cp5.getController("Direction")).setOff();
		} else if (jSet.getString("Learning").equals("Inertial")) {
			((Button)cp5.getController("Inertial")).setOn();
			((Button)cp5.getController("Random")).setOff();
			((Button)cp5.getController("Direction")).setOff();
		} else if (jSet.getString("Learning").equals("Direction")) {
			((Button)cp5.getController("Direction")).setOn();
			((Button)cp5.getController("Random")).setOff();
			((Button)cp5.getController("Inertial")).setOff();
		}
		
		for (int a = 0; a < (int) cp5.getController("Num_Cultures").getValue(); a++) {
			JSONObject jc_a = jso.getJSONObject(a + "_settings");
			JSONObject jc_b = jSet.getJSONObject(a +"_settings");
			
			for (String s : Culture.CultureControls) {
				JSONObject jss = jc_b.getJSONObject(s);
				
				if (!jss.getBoolean("Override") && !jSet.getBoolean("Global_Override")) {
					if (jss.getBoolean("Range")) {
						float min = jss.getFloat("min");
						float max = jss.getFloat("max");
						float minV = ((Range)cp5.getController(a + "_" + s + "_r")).getMin();
						float maxV = ((Range)cp5.getController(a + "_" + s + "_r")).getMax();
						
						ReplaceRange((char)(a+'0'), s, minV, maxV, min, max);
						cp5.getController(a + "_" + s).hide();
						((Button)cp5.getController(a + "_" + s + "_b")).setOn();
						((Button)cp5.getController(a + "_" + s + "_a")).setOn();
						setActiveParameter(a + "_" + s + "_a");
					} else {
						cp5.getController(a + "_" + s).show();
						cp5.getController(a + "_" + s).setValue( jss.getFloat( "value" ) );
						((Button)cp5.getController(a + "_" + s + "_b")).setOff();
						cp5.getController(a + "_" +s+"_r").hide();
					}
				} else {
					if (a > 0) {
						cp5.getController(a + "_" +s+"_r").hide();
						cp5.getController(a + "_" +s).hide();
					}
					if (a==0) {
						if (jss.getBoolean("Range")) {
							float min = jss.getFloat("min");
							float max = jss.getFloat("max");
							float minV = ((Range)cp5.getController(a + "_" + s + "_r")).getMin();
							float maxV = ((Range)cp5.getController(a + "_" + s + "_r")).getMax();
							
							ReplaceRange((char)(a+'0'), s, minV, maxV, min, max);
							cp5.getController(a + "_" + s).hide();
							((Button)cp5.getController(a + "_" + s + "_b")).setOn();
							((Button)cp5.getController(a + "_" + s + "_a")).setOn();
							setActiveParameter(a + "_" + s + "_a");
						} else {
							cp5.getController(a + "_" + s).show();
							cp5.getController(a + "_" + s).setValue( jss.getFloat( "value" ) );
							((Button)cp5.getController(a + "_" + s + "_b")).setOff();
							cp5.getController(a + "_" +s+"_r").hide();
						}
						
						// if not global override
						if (jss.getBoolean("Override"))
							((Button)cp5.getController(a + "_" + s + "_x" )).setOn();
					}
					
				}
			}
		}

		//loadView = false;
		for (Controller cc : cp5Main) {
			cc.setVisible(false);
		}
		UpdateUILabels();
		expiredSolution = false;
	}
	
	public void setActiveParameter (String s) {
		println(s);
		
		
		if (!aParams[0].equals(s) && !aParams[1].equals(s)) {
			if (!aParams[1].equals("")) 
				((Button)cp5.getController(aParams[1])).setOff();

			String st = aParams[0];
			aParams[1] = st;
			aParams[0] = s;
		} else {
			if (aParams[0].equals(s)) {
				aParams[0] = "";										
			}
			if (aParams[1].equals(s)) {
				aParams[1] = "";										
			}
		}
	}
	
	public int getNumRanges() {
		int num = 0;
		
		boolean c = ((Button) cp5.getController("Global_Override")).isOn();
		
		for (int i = 0; i < (int) cp5.getController("Num_Cultures").getValue(); i++) {
			if (i == 0 || !c) {
				for (String s : Culture.CultureControls) {
					if ( ((Button)cp5.getController( i + "_" + s + "_b")).isOn() && (i == 0 || !((Button)cp5.getController( "0_" + s + "_x")).isOn()   ) ) {
						num++;
					}
				}
			}
		}
		return num;
	}
		
	public void keyPressed() {
		if (key =='z') {
			saveFrame("r-######.jpg");
		}
	}

	public JSONObject getGSettings() {
		/* distinct from the settings that are stored within each iteration, this will store the ranges for parameters that are shifting etc */
		
		JSONObject jsSettings = new JSONObject();					// stores all of the settings we will use
		jsSettings.setInt("Num_Iterations", (int) cp5.getController("Num_Iterations").getValue());
		jsSettings.setInt("Num_Cultures", (int) cp5.getController("Num_Cultures").getValue());
		jsSettings.setInt("World_Size", (int) cp5.getController("World_Size").getValue());
		jsSettings.setInt("Time_Steps", (int) cp5.getController("Time_Steps").getValue());
		jsSettings.setInt("Num_Steps", (int) cp5.getController("Num_Steps").getValue());
		jsSettings.setDouble("Random_Moves", cp5.getController("Random_Moves").getValue());
		jsSettings.setBoolean("VN", ((Button)cp5.getController("VN")).isOn());
		jsSettings.setBoolean("Venue_Lock", ((Button)cp5.getController("Venue_Lock")).isOn());
		jsSettings.setBoolean("Distribution_Lock", ((Button)cp5.getController("Distribution_Lock")).isOn());
		jsSettings.setBoolean("Global_Override", ((Button)cp5.getController("Global_Override")).isOn());
		
		// learning
		if ( ((Button)cp5.getController("Random")).isOn() ) {
			jsSettings.setString("Learning", "Random");
		} else if ( ((Button)cp5.getController("Inertial")).isOn() ) {
			jsSettings.setString("Learning", "Inertial");
		} else if ( ((Button)cp5.getController("Direction")).isOn() ) {
			jsSettings.setString("Learning", "Direction");
		}
		// movement type
		if ( ((Button)cp5.getController("Best")).isOn()  ) {
			jsSettings.setString("Movement", "Best");
		} else if ( ((Button)cp5.getController("Closest")).isOn()  ) {
			jsSettings.setString("Movement", "Closest");
		} else if ( ((Button)cp5.getController("Rand_Works")).isOn()  ) {
			jsSettings.setString("Movement", "Rand");
		}
		// above is duplicated from below
		
		// these are the cultural settings
		for (int a = 0; a < (int) cp5.getController("Num_Cultures").getValue(); a++) {
			JSONObject jsCultureSettings = new JSONObject();
			jsCultureSettings.setInt("id", a);


			for (int j = 0; j < pNames.length; j++) {
				JSONObject oneSetting = new JSONObject();
				
				String axis = "";
				if (aParams[0].equals(a+"_"+pNames[j]+"_a")) axis = "x";
				if (aParams[1].equals(a+"_"+pNames[j]+"_a")) axis = "y";
				
				boolean override =  ((Button)cp5.getController( "0_"+pNames[j]+"_x")).isOn();
				boolean range = ((Button)cp5.getController( a+ "_"+pNames[j]+"_b")).isOn() 
										|| (override && ((Button)cp5.getController( "0_"+pNames[j]+"_b")).isOn());
				
				Range rr = (Range) cp5.getController("0_"+pNames[j]+"_r");
				if (!override)
					rr = (Range) cp5.getController(a + "_"+pNames[j]+"_r");
				
				double val = cp5.getController("0_"+pNames[j]).getValue();
				if (!override)
					val = cp5.getController(a + "_"+pNames[j]).getValue();
				
				oneSetting.setBoolean("Override", override);
				oneSetting.setBoolean("Range", range);
				oneSetting.setDouble("value", range ? 0 : val );
				oneSetting.setDouble("min", range ? rr.getLowValue() : 0);
				oneSetting.setDouble("max", range ? rr.getHighValue() : 0);
				oneSetting.setString("axis", axis);
				
				jsCultureSettings.setJSONObject(pNames[j], oneSetting);
			}
			jsSettings.setJSONObject(a + "_settings", jsCultureSettings);
		}
		return jsSettings;
	}
	
	// the settings of a single iteration
	public JSONObject GenerateSettings(int numSteps, int numRanges, int iter, int z) {
		JSONObject jsSettings = new JSONObject();					// stores all of the settings we will use
		jsSettings.setInt("Num_Iterations", (int) cp5.getController("Num_Iterations").getValue());
		jsSettings.setInt("Num_Cultures", (int) cp5.getController("Num_Cultures").getValue());
		jsSettings.setInt("World_Size", (int) cp5.getController("World_Size").getValue());
		jsSettings.setInt("Time_Steps", (int) cp5.getController("Time_Steps").getValue());
		jsSettings.setInt("Num_Steps", (int) cp5.getController("Num_Steps").getValue());
		jsSettings.setDouble("Random_Moves", cp5.getController("Random_Moves").getValue());
		jsSettings.setBoolean("VN", ((Button)cp5.getController("VN")).isOn());
		jsSettings.setBoolean("Venue_Lock", ((Button)cp5.getController("Venue_Lock")).isOn());
		jsSettings.setBoolean("Distribution_Lock", ((Button)cp5.getController("Distribution_Lock")).isOn());
		jsSettings.setBoolean("Global_Override", ((Button)cp5.getController("Global_Override")).isOn());
		
		// learning
		if ( ((Button)cp5.getController("Random")).isOn() ) {
			jsSettings.setString("Learning", "Random");
		} else if ( ((Button)cp5.getController("Inertial")).isOn() ) {
			jsSettings.setString("Learning", "Inertial");
		} else if ( ((Button)cp5.getController("Direction")).isOn() ) {
			jsSettings.setString("Learning", "Direction");
		}
		// movement type
		if ( ((Button)cp5.getController("Best")).isOn()  ) {
			jsSettings.setString("Movement", "Best");
		} else if ( ((Button)cp5.getController("Closest")).isOn()  ) {
			jsSettings.setString("Movement", "Closest");
		} else if ( ((Button)cp5.getController("Rand_Works")).isOn()  ) {
			jsSettings.setString("Movement", "Rand");
		}
		
		
//		int countUpRanges = 0;
		boolean gOverride = ((Button)cp5.getController("Global_Override")).isOn();
		
		
		// these are the cultural settings
		for (int a = 0; a < (int) cp5.getController("Num_Cultures").getValue(); a++) {
			JSONObject jsCultureSettings = new JSONObject();
			jsCultureSettings.setInt("id", a);

			for (int j = 0; j < pNames.length; j++) {
				// three options 
				// 1 -> global override is on, assign the value of culture 0 for all cultures 0 to a
				// 2 -> local override is on, assign the value of culture 0 for all cultures 0 to a
				// 3 -> no override, assign a value based on which parameter/dimension this is
				
				// z is the current iteration
				// iter - num. iterations per step
				// numSteps
				// numRanges
				
				// 1 or 2
				if (gOverride || ((Button)cp5.getController( "0_"+pNames[j]+"_x")).isOn() ) {
					// driving parameter or copied parameter
					if (a == 0) {
						// range value or single value
						if (  ((Button)cp5.getController( "0_"+pNames[j]+"_b")).isOn() ) {
							Range rr = (Range) cp5.getController("0_"+pNames[j]+"_r");
							
							int dimF = 0;
							if (aParams[0].equals("0_"+pNames[j]+"_a")) dimF = 0;
							if (aParams[1].equals("0_"+pNames[j]+"_a")) dimF = 1;
							
							int val = floor(z / (iter * pow(numSteps,dimF))) % numSteps;
							float valSet = rr.getLowValue() + (rr.getHighValue() - rr.getLowValue()) *   (numSteps==1 ? 1 : (val / ((float) numSteps-1)));
							jsCultureSettings.setDouble(pNames[j], valSet);	
						} else {
							jsCultureSettings.setDouble(pNames[j], cp5.getController("0_"+pNames[j]).getValue());	
						}
					} else {
						jsCultureSettings.setDouble(pNames[j], jsSettings.getJSONObject("0_settings").getDouble(pNames[j]));	// copy from 0
					}
				} else {
					if (  ((Button)cp5.getController( a+"_"+pNames[j]+"_b")).isOn() ) {
						Range rr = (Range) cp5.getController(a+"_"+pNames[j]+"_r");
						
						int dimF = 0;
						if (aParams[0].equals(a+"_"+pNames[j]+"_a")) dimF = 0;
						if (aParams[1].equals(a+"_"+pNames[j]+"_a")) dimF = 1;
						
						int val = floor(z / (iter * pow(numSteps,dimF))) % numSteps;
						float valSet = rr.getLowValue() + (rr.getHighValue() - rr.getLowValue()) *   (numSteps==1 ? 1 : (val / ((float) numSteps-1)));
						jsCultureSettings.setDouble(pNames[j], valSet);	
					} else {
						jsCultureSettings.setDouble(pNames[j], cp5.getController(a+"_"+pNames[j]).getValue());	
					}
				}
			}
			
			String ven = "";
			if ( ((Button) cp5.getController("Venue_Lock")).isOn() ) {
				for (PVector pv : sr[ai].VenueLocations) {

					if (pv.z == a) {
						ven += pv.x+"&"+pv.y+"&";
					}
				}
			}
			String hu = "";
			if ( ((Button) cp5.getController("Distribution_Lock")).isOn() ) {
				for ( Agent_Human ha : sr[ai].ha  ) {
					if (ha.cultureIndex == a) {
						hu += ha.store_pos[time][0]+"&"+ha.store_pos[time][1]+"&";
					}
				}
			}
			jsCultureSettings.setString("VenLocations", ven);
			jsCultureSettings.setString("AgentLocations", hu);
			jsSettings.setJSONObject(a + "_settings", jsCultureSettings);
		}
		
		return jsSettings;
	}
	
	public JSONObject getGAnalysis(SingleRun[] sRuns) {
		JSONObject gA = new JSONObject();
		
		String xAxis = "";
		PVector xR = new PVector(0,0);
		String yAxis = "";
		PVector yR = new PVector(0,0);
		
		// 
		if (!aParams[0].equals("")) {
			xAxis = aParams[0];
			Range xRange = (Range) cp5.getController( aParams[0].substring(0,aParams[0].length()-1) + "r" );
			xR = new PVector ( xRange.getLowValue(), xRange.getHighValue() );
		}
		
		if (!aParams[1].equals("")) {
			yAxis = aParams[1];
			Range yRange = (Range) cp5.getController( aParams[1].substring(0,aParams[1].length()-1) + "r" );
			yR = new PVector ( yRange.getLowValue(), yRange.getHighValue() );
		}
		int numStepsX = xAxis.equals("") ? 1 : (int) cp5.getController("Num_Steps").getValue();
		int numStepsY = yAxis.equals("") ? 1 : (int) cp5.getController("Num_Steps").getValue();
		
		mpa = new float[numStepsX][numStepsY];
		
		JSONArray gDiversityT = new JSONArray();
		JSONArray gUnsettledT = new JSONArray();
		JSONArray gVenFracT = new JSONArray();
		JSONArray gNumClustersT = new JSONArray();
		JSONArray gPerimAreaT = new JSONArray();
		JSONArray gMStabilityT = new JSONArray();
		

		for (int ti = 0; ti < maxTime; ti++) {
			
			JSONArray gDiversity = new JSONArray();
			JSONArray gUnsettled = new JSONArray();
			JSONArray gVenFrac = new JSONArray();
			JSONArray gNumClusters = new JSONArray();
			JSONArray gPerimArea = new JSONArray();
			JSONArray gMStability = new JSONArray();
			
			for (int i = 0; i < numStepsX; i++) {
				float xVal = 0;	
				xVal = xR.x + (xR.y-xR.x) * (i/(float) (numStepsX-1));

				for (int j = 0; j < numStepsY; j++) {
					float yVal = 0;
					yVal = yR.x + (yR.y-yR.x) * (  (numStepsY-1-j)       /(float)(numStepsY-1));

					int num = 0;
					float tot0=0;
					float tot1=0;
					float tot2=0;
					float tot3=0;
					float tot4=0;
					float tot5=0;

					for (SingleRun sr : sRuns) {
						if ( (xAxis.equals("") || Math.abs( sr.jsP.getJSONObject(xAxis.charAt(0)+"_settings").getFloat( xAxis.split("_")[1]) - xVal) < 0.001) 
								&& (yAxis.equals("") || Math.abs( sr.jsP.getJSONObject(yAxis.charAt(0)+"_settings").getFloat( yAxis.split("_")[1]) - yVal) < 0.001 )) {
							num++;

							tot0 += sr.DiversityPerTime[ti];
							tot1 += sr.UnsettledPerTime[ti];
							tot2 += sr.VenFractionPerTime[ti];
							tot3 += sr.numClusters[ti];
							tot4 += sr.perimArea[ti];
							tot5 += sr.mStability[ti];
						}
					}
					tot0 = tot0 / (float) num;
					tot1 = tot1 / (float) num;
					tot2 = tot2 / (float) num;
					tot3 = tot3 / (float) num;
					tot4 = tot4 / (float) num;
					tot5 = tot5 / (float) num;
					
					gDiversity.append(tot0);
					gUnsettled.append(tot1);
					gVenFrac.append(tot2);
					gNumClusters.append(tot3);
					gPerimArea.append(tot4);
					gMStability.append(tot5);
				}
			}
			
			gDiversityT.append(gDiversity);
			gUnsettledT.append(gUnsettled);
			gVenFracT.append(gVenFrac);
			gNumClustersT.append(gNumClusters);
			gPerimAreaT.append(gPerimArea);
			gMStabilityT.append(gMStability);
		}
		
		gA.setJSONArray("Diversity", gDiversityT);
		gA.setJSONArray("Unsettled", gUnsettledT);
		gA.setJSONArray("VenFrac", gVenFracT);
		gA.setJSONArray("NumClusters", gNumClustersT);
		gA.setJSONArray("PerimArea", gPerimAreaT);
		gA.setJSONArray("mStability", gMStabilityT);
		
		return gA;
	}
	
	
	public void UpdateUILabels() {
		// go through all of the ui, change the caption labels back to their default, unless this particular active run has a range..
		for (int a = 0; a < (int) cp5.getController("Num_Cultures").getValue(); a++) {

			for (String s : Culture.CultureControls) {
				if ( ((Button)cp5.getController(a + "_"+ s +"_b")).isOn() ) {
					String sm = cp5.getController(a+"_"+s+"_r").getCaptionLabel().getText() ;
					if (sm.indexOf("-")!=-1) {
						sm = sm.substring(0,sm.indexOf("-")-1);
					}
					sm += " - " + String.format("%.2f", sr[ai].jsP.getJSONObject(a+"_settings").getDouble(s) );
					cp5.getController(a+"_"+s+"_r").setCaptionLabel(sm);
				} else {
					String sm = cp5.getController(a+"_"+s+"_r").getCaptionLabel().getText() ;
					if (sm.indexOf("-")!=-1) {
						sm = sm.substring(0,sm.indexOf("-")-1);
					}
				}
			}
		}	

			
	}
	
	public void OverrideSingle(String na, boolean on) {
		// anything that is being hidden or revealed, we should double check that it has the correct value, ie. the value of its corresponding partner
		
		float nc = cp5.getController("Num_Cultures").getValue();
		
		for (int i = 1; i < 5; i++) {
			cp5.getController(i + "_" + na + "_a").setVisible(!on && i < nc);
			cp5.getController(i + "_" + na + "_b").setVisible(!on && i < nc);
			
			if ( ((Button)cp5.getController(i + "_" + na + "_b")).isOn()  ) {
				cp5.getController(i + "_" + na + "").setVisible(false);
				cp5.getController(i + "_" + na + "_r").setVisible( !on && i < nc );
			} else {
				cp5.getController(i + "_" + na + "").setVisible(!on && i < nc);
				cp5.getController(i + "_" + na + "_r").setVisible( false );
			}
			cp5.getController(i + "_" + na + "_t").setVisible(!on && i < nc);
		}
	}
	
	public void ReplaceRange(char ind, String pName, double minVal, double maxVal, float v1, float v2 ) {
		println(ind +"_" + pName);
		float sliX = cp5.getController(ind + "_" + pName + "_r").getPosition()[0];
		float sliY = cp5.getController(ind + "_" + pName + "_r").getPosition()[1];
		int w = cp5.getController(ind + "_" + pName + "_r").getWidth();
		int k = cp5.getController(ind + "_" + pName + "_r").getHeight();
		
		cp5Main.remove( cp5.getController(ind + "_" + pName + "_r") );
		cp5.getController(ind + "_" + pName + "_r").remove();
		cp5.getController(ind + "_" + pName ).setVisible(false);
		
		cp5.addRange(ind +"_" + pName + "_r")
		.setMin((float) minVal)
		.setMax((float) maxVal )
		.setLowValue( v1 )
		.setHighValue( v2 )
		.setPosition(sliX, sliY)
		.setSize(w,k)
		.onChange( new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				String stz = theEvent.getController().getName().split("_")[1];
				
				if (theEvent.getController().getName().charAt(0)=='0' && ((Button)cp5.getController("0_"+stz+"_x")).isOn() ) {
					Range rr = (Range) theEvent.getController();
					
					for (int i = 1; i < 5; i++) {
						((Range)cp5.getController(i+"_"+stz+"_r")).setLowValue( rr.getLowValue() );
						((Range)cp5.getController(i+"_"+stz+"_r")).setHighValue( rr.getHighValue() );
					}
				}
			}
		})
		.setHandleSize(4)
		.setCaptionLabel(pName)									
		.setColorCaptionLabel(cp5.BLACK)
		.getCaptionLabel().align(ControlP5.LEFT_OUTSIDE, CENTER);
		
		cp5Main.add(cp5.getController(ind + "_" + pName + "_r"));
		
	}
	
	public void AddCultureParameter (int i, String pName, double minVal, double maxVal, double initialVal, int y , int sliX, int texX, int butX, int actX, int w, int k, int tw) {
		
		/* SLIDER */
		cp5.addSlider(i+"_" + pName)
		.setRange((float)minVal,(float)maxVal)
		.setValue((float) initialVal)
		.setPosition(sliX,y)
		.setSize(w,k)
		.setVisible( i < cp5.getController("Num_Cultures").getValue() )
		.onChange(new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				if (theEvent.getController().getName().charAt(0)=='0' && ((Button)cp5.getController("0_" + pName + "_x")).isOn() ) {
					for (int i = 1; i < 5; i++) {
						cp5.getController(i+"_"+pName).setValue(theEvent.getController().getValue());
					}
				}
			}}
		)
		.setCaptionLabel(pName)
		.setColorCaptionLabel(cp5.BLACK)
		.getCaptionLabel().align(ControlP5.LEFT_OUTSIDE, CENTER)
		;
		cp5Main.add(cp5.getController(i + "_" + pName));

		/* TEXT */
		Controller c = cp5.addTextfield(i + "_"+pName+"_t")
		.setValue(2)
		.setPosition(texX,y)
		.setSize(tw-10,k)
		.setVisible( i < cp5.getController("Num_Cultures").getValue() )
		.onChange( new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				String stxt = ((Textfield)theEvent.getController()).getText();
				char ind = theEvent.getController().getName().charAt(0);
				
				if (stxt.indexOf(":") != -1) {
					String s1 = stxt.split(":")[0];
					String s2 = stxt.split(":")[1];
					float v1 = Float.parseFloat(s1);
					float v2 = Float.parseFloat(s2);
					
					((Button) cp5.getController(ind + "_" + pName + "_b")).setOn();
					ReplaceRange(ind, pName, minVal, maxVal, v1, v2 );
					
					String szt = theEvent.getController().getName().split("_")[1];
					if (ind=='0' && ((Button)cp5.getController("0_"+szt+"_x")).isOn()  ) {
						for (int i = 1; i < 5; i++) {
							((Button) cp5.getController(i + "_"+szt+"_b")).setOn();
							ReplaceRange((char)(i+'0'), pName, minVal, maxVal, v1, v2 );		// pName for szt?
							cp5.getController((char)(i+'0') + "_"+pName+"_r").setVisible(false);
						}
					}
				} else {
					((Button) cp5.getController(ind + "_" + pName + "_b")).setOff();
					float v = Float.parseFloat(stxt);
					cp5.getController(ind + "_" + pName ).setValue(v);
					
					String szt = theEvent.getController().getName().split("_")[1];
					if (ind=='0' && ((Button)cp5.getController("0_"+szt+"_x")).isOn() ) {
						for (int i = 1; i < 5; i++) {
							cp5.getController(i+"_"+szt).setValue(v);
							((Button) cp5.getController(i + "_"+szt+"_b")).setOff();
						}
					}
				}
			}
		});
		cp5Main.add(c);

		/* RANGE */
		cp5.addRange(i +"_" + pName + "_r")
		.setMin((float)minVal)
		.setMax((float)maxVal)
		.setLowValue( (float)(minVal+initialVal) / 2f)
		.setHighValue((float)(maxVal+initialVal) / 2f)
		.setPosition(sliX,y)
		.setSize(w,k)
		.setHandleSize(4)
		.onChange( new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				String stz = theEvent.getController().getName().split("_")[1];
				
				if (theEvent.getController().getName().charAt(0)=='0' && ((Button)cp5.getController("0_"+stz+"_x")).isOn() ) {
					Range rr = (Range) theEvent.getController();
					
					for (int i = 1; i < 5; i++) {
						((Range)cp5.getController(i+"_"+stz+"_r")).setLowValue( rr.getLowValue() );
						((Range)cp5.getController(i+"_"+stz+"_r")).setHighValue( rr.getHighValue() );
					}
				}
			}
		})
		.setVisible( false )
		.setCaptionLabel(pName)
		.setColorCaptionLabel(cp5.BLACK)
		.getCaptionLabel().align(ControlP5.LEFT_OUTSIDE, CENTER);
		
		cp5Main.add(cp5.getController(i + "_"+pName+"_r"));			
		
		
		c = cp5.addButton(i + "_"+pName+"_b")
		.setPosition(butX,y)
		.setSize(tw-18,k)
		.setSwitch(true)
		.setCaptionLabel("")
		.setVisible( i < cp5.getController("Num_Cultures").getValue() )
		.onClick(new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				if ( ((Button)theEvent.getController()).getBooleanValue() ) {
					cp5.getController(((Button)theEvent.getController()).getName().charAt(0) + "_"+pName).setVisible(false);
					cp5.getController(((Button)theEvent.getController()).getName().charAt(0) + "_"+pName+"_r").setVisible(true);
				} else {
					cp5.getController(((Button)theEvent.getController()).getName().charAt(0) + "_"+pName).setVisible(true);
					cp5.getController(((Button)theEvent.getController()).getName().charAt(0) + "_"+pName+"_r").setVisible(false);
				}
				
				String stz = theEvent.getController().getName().split("_")[1];
				if (theEvent.getController().getName().charAt(0)=='0' && ((Button)cp5.getController("0_"+stz+"_x")).isOn() ) {
					for (int i = 1; i < 5; i++) {
						if (((Button)theEvent.getController()).isOn()) {
							((Button)cp5.getController(i+"_"+stz+"_b")).setOn();
						} else {
							((Button)cp5.getController(i+"_"+stz+"_b")).setOff();
						}
					}
				}
			}
		});
		cp5Main.add(c);
		
		/* ACTIVE DIMENSION */
		c = cp5.addButton(i + "_"+pName+"_a")
			.setPosition(actX,y)
			.setSize(tw-18,k)
			.setSwitch(true)
			.setCaptionLabel("")
			.setVisible( i < cp5.getController("Num_Cultures").getValue() )
			.onClick(new CallbackListener() {
				public void controlEvent(CallbackEvent theEvent) {
					String s = theEvent.getController().getName();

					setActiveParameter(s);
				}
			});
			cp5Main.add(c);
		
		/* OVERRIDE */
		if (i==0) {
			c = cp5.addButton(i + "_"+pName+"_x")
			.setPosition(actX+15,y)
			.setSize(tw-18,k)
			.setSwitch(true)
			.setCaptionLabel("")
			.onClick(new CallbackListener() { 
				public void controlEvent(CallbackEvent theEvent) {
					OverrideSingle(theEvent.getController().getName().split("_")[1], ((Button) theEvent.getController()).isOn());
				}
			});
			cp5Main.add(c);		
		}
	}
	
	
	
	public void addUI() {
		cp5 = new ControlP5(this);
		cp5.setAutoDraw(false);
		
		cp5.setColorBackground(cp5.GRAY);
		cp5.setColorForeground(cp5.BLACK);
		cp5.setColorActive(cp5.BLACK);
		cp5.setColorValueLabel(cp5.WHITE);
		
		int k = 9;
		int k2 = 10;
		int w = 60;
		int tw = 30;

		cp5.addButton("WTitle")
		.setPosition(gWPos.x,gWPos.y-k)
		.setSize(150,k)
		.setCaptionLabel("World");
		
		cp5.addButton("ATitle")
		.setPosition(gAPos.x,gAPos.y-k)
		.setSize(150,k)
		.setCaptionLabel("Analysis - " + AnalysisTypes[aT])
		.onPress(new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				aT = (aT+1) % (AnalysisTypes.length);
				theEvent.getController().setCaptionLabel( "Analysis - " + AnalysisTypes[aT] );
				cp5.getController("Vis_Type").setCaptionLabel( vNames[aT][vT[aT]] );
				
				redrawViews = true;
				DrawAnalysis();
				// Redraw the analysis view
			}
		});
		cp5.addButton("Vis_Type")
		.setPosition(gAPos.x+155,gAPos.y-k)
		.setSize(349,k)
		.setCaptionLabel("")
		.onPress(new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				vT[aT]++;
				if (vT[aT] > vNames[aT].length-1)
					vT[aT] = 0;
				theEvent.getController().setCaptionLabel(vNames[aT][vT[aT]]);
				
				redrawViews = true;
			}
		});
		
		
		
		cp5.addButton("VTitle")
		.setPosition(gFPos.x,gFPos.y-k)
		.setSize(150,k)
		.setCaptionLabel("Form Layout - Venues");
		cp5.addButton("DTitle")
		.setPosition(gDPos.x,gDPos.y-k)
		.setSize(150,k)
		.setCaptionLabel("Initial Distribution");
		
		
		// need to also add a textbox option
		cp5.addSlider("Num_Iterations")
		.setRange(1,100)
		.setValue(1)
		.setPosition(width-w-tw,k2+=10)
		.setSize(w,k)
		.onChange(new CallbackListener() {
		      public void controlEvent(CallbackEvent theEvent) {
		    	  theEvent.getController().changeValue( floor( theEvent.getController().getValue()) );
	//	    	  ((Textfield)cp5.getController("Num_Iterations_t")).setText( floor( theEvent.getController().getValue()) + " " );
		    	  ai = 0;
		    	  if (cp5.getController("it")!=null)
		    		  cp5.getController("it").setCaptionLabel(ai+" / "+sr.length);
		    	  srNew = null;
		    	  expiredSolution = true;
		      }})
		.setCaptionLabel("Num Iterations")
		.setColorCaptionLabel(cp5.BLACK)
		.getCaptionLabel().align(ControlP5.LEFT_OUTSIDE, CENTER);
		;
		cp5Main.add( cp5.getController("Num_Iterations") );
		
		Controller c = cp5.addButton("it")
				.setValue(0)
				.setPosition(380,1)
				.setSize(39,k)
				.setCaptionLabel(ai+" / "+(sr.length))
				.onPress(new CallbackListener() {
					public void controlEvent(CallbackEvent theEvent) {
						int aai = ai;
						
						ai = 0;
						setIteration(aai,ai);
					}
				});
		cp5Main.add(c);
		
		c = cp5.addButton(">a")
	     .setValue(0)
	     .setPosition(420,1)
	     .setSize(19,k)
	     .onPress(new CallbackListener() {
		      public void controlEvent(CallbackEvent theEvent) {
		    	  int aai = ai;
		    	  
		    	  ai++;
		    	  if (ai > sr.length-1) ai = 0;
		    	  	setIteration(aai, ai);
		      }
	     })
	     .setCaptionLabel(">");
		cp5Main.add(c);
		
		c = cp5.addButton("<a")
	     .setValue(0)
	     .setPosition(360,1)
	     .setSize(19,k)
	     .onPress(new CallbackListener() {
		      public void controlEvent(CallbackEvent theEvent) {
		    	int aai = ai;
		    	ai--;
				if (ai < 0) ai = sr.length-1;
				
				setIteration(aai, ai);
		      }
	     })
	     .setCaptionLabel("<");
		cp5Main.add(c);
		
		
		
		
		c = cp5.addTextfield("Num_Iterations_t")
		.setValue(1)
		.setPosition(width-tw+5,k2)
		.setSize(tw-5,k)
		.onChange( new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				float v = Float.parseFloat(((Textfield)theEvent.getController()).getText());
				cp5.getController("Num_Iterations").setValue(v);
			}
		});
		cp5Main.add(c);
				
		cp5.addSlider("Num_Steps")
		.setRange(1,100)
		.setValue(1)
		.setPosition(width-w-tw,k2+=10)
		.setSize(w,k)
		.onChange(new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				theEvent.getController().changeValue( floor( theEvent.getController().getValue()) );

			    srNew = null;
			    expiredSolution = true;
			}})
		.setCaptionLabel("Num Steps")
		.setColorCaptionLabel(cp5.BLACK)
		.getCaptionLabel().align(ControlP5.LEFT_OUTSIDE, CENTER);
		;
		cp5Main.add( cp5.getController("Num_Steps") );
		
		c = cp5.addTextfield("Num_Steps_t")
				.setValue(1)
				.setPosition(width-tw+5,k2)
				.setSize(tw-5,k)
				.onChange( new CallbackListener() {
					public void controlEvent(CallbackEvent theEvent) {
						float v = Float.parseFloat(((Textfield)theEvent.getController()).getText());
						cp5.getController("Num_Steps").setValue(v);
					}
				});
				cp5Main.add(c);
		
				
		cp5.addSlider("Num_Cultures")
		.setRange(2,5)
		.setValue(2)
		.setPosition(width-w-tw,k2+=10)
		.setSize(w,k)
		.setCaptionLabel("Num Cultures")
		.setColorCaptionLabel(cp5.BLACK)
		.getCaptionLabel().align(ControlP5.LEFT_OUTSIDE, CENTER);
		;
		cp5Main.add(cp5.getController("Num_Cultures"));
		
		c = cp5.addTextfield("Num_Cultures_t")
		.setValue(2)
		.setPosition(width-tw+5,k2)
		.setSize(tw-5,k)
		.onChange( new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				float v = Float.parseFloat(((Textfield)theEvent.getController()).getText());
				cp5.getController("Num_Cultures").setValue(v);
			}
		});
		cp5Main.add(c);

		cp5.addSlider("World_Size")
		.setRange(10,200)
		.setValue(50)
		.setPosition(width-w-tw,k2+=10)
		.setSize(w,k)
		.setCaptionLabel("World Size")
		.setColorCaptionLabel(cp5.BLACK)
		.getCaptionLabel().align(ControlP5.LEFT_OUTSIDE, CENTER);
		;
		cp5Main.add(cp5.getController("World_Size"));
		
		c = cp5.addTextfield("World_Size_t")
		.setValue(50)
		.setPosition(width-tw+5,k2)
		.setSize(tw-5,k)
		.onChange( new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				float v = Float.parseFloat(((Textfield)theEvent.getController()).getText());
				cp5.getController("World_Size").setValue(v);
			}
		});
		cp5Main.add(c);
		
		
		/* big addition */
		cp5.addSlider("Time_Steps")
		.setRange(5,25000)
		.setValue(20)
		.setPosition(width-w-tw,k2+=10)
		.setSize(w,k)
		.setCaptionLabel("Time Steps")
		.setColorCaptionLabel(cp5.BLACK)
		.getCaptionLabel().align(ControlP5.LEFT_OUTSIDE, CENTER);
		;
		cp5Main.add(cp5.getController("Time_Steps"));
		
		c = cp5.addTextfield("Time_Steps_t")
		.setValue(20)
		.setPosition(width-tw+5,k2)
		.setSize(tw-5,k)
		.onChange( new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				float v = Float.parseFloat(((Textfield)theEvent.getController()).getText());
				cp5.getController("Time_Steps").setValue(v);
			}
		});
		cp5Main.add(c);

		cp5.addSlider("Random_Moves")
		.setRange(0,1f)
		.setValue(0.01f)
		.setPosition(width-w-tw,k2+=10)
		.setSize(w,k)
		.setCaptionLabel("Random Movement")
		.setColorCaptionLabel(cp5.BLACK)
		.getCaptionLabel().align(ControlP5.LEFT_OUTSIDE, CENTER);
		;
		cp5Main.add(cp5.getController("Random_Moves"));
		
		c = cp5.addTextfield("Random_Moves_t")
		.setValue(0.001f)
		.setPosition(width-tw+5,k2)
		.setSize(tw-5,k)
		.onChange( new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				float v = Float.parseFloat(((Textfield)theEvent.getController()).getText());
				cp5.getController("Random_Moves").setValue(v);
			}
		});
		cp5Main.add(c);

		
		
		
		// global culture overrides
		c = cp5.addButton("Venue_Lock")
		.setPosition(width-w,k2+=10)
		.setSize(w,k)
		.setSwitch(true)
		.setCaptionLabel("Venue Lock")
		.onPress(new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				
			}
		});
		cp5Main.add(c);

		// global culture overrides
		c = cp5.addButton("Distribution_Lock")
		.setPosition(width-w*2-1,k2)
		.setSize(w,k)
		.setSwitch(true)
		.setCaptionLabel("Distr. Lock")
		.onPress(new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
			}
		});
		cp5Main.add(c);
 
		c = cp5.addButton("VN")
				.setPosition(width-w*2-1,k2+=10)
				.setSize(w,k)
				.setSwitch(true)
				.setOn()
				.setCaptionLabel("Von Neumann")
				.onClick(new CallbackListener() {
					public void controlEvent(CallbackEvent theEvent) {
						if ( ((Button)theEvent.getController()).getBooleanValue() ) {
							((Button)cp5.getController("M")).setOff();
							
						} else {
							((Button)cp5.getController("M")).setOn();
						}
					}
				});
		cp5Main.add(c);

		c = cp5.addButton("M")
				.setPosition(width-w,k2)
				.setSize(w,k)
				.setSwitch(true)
				.setCaptionLabel("Moore")
				.onClick(new CallbackListener() {
					public void controlEvent(CallbackEvent theEvent) {
						if (  ((Button)theEvent.getController()).getBooleanValue() ) {
							((Button)cp5.getController("VN")).setOff();
						} else {
							((Button)cp5.getController("VN")).setOn();
						}
					}
				});
		cp5Main.add(c);
		
		c = cp5.addButton("Random")
				.setPosition(width-w*3-2,k2+=10)
				.setSize(w,k)
				.setSwitch(true)
				.setCaptionLabel("Random")
				.onClick(new CallbackListener() {
					public void controlEvent(CallbackEvent theEvent) {
						if ( ((Button)theEvent.getController()).getBooleanValue() ) {
							((Button)cp5.getController("Inertial")).setOff();
							((Button)cp5.getController("Direction")).setOff();
						} else {
							((Button)cp5.getController("Random")).setOn();
						}
					}
				});
		cp5Main.add(c);
		
		c = cp5.addButton("Inertial")
				.setPosition(width-w*2-1,k2)
				.setSize(w,k)
				.setSwitch(true)
				.setOn()
				.setCaptionLabel("Inertial")
				.onClick(new CallbackListener() {
					public void controlEvent(CallbackEvent theEvent) {
						if ( ((Button)theEvent.getController()).getBooleanValue() ) {
							((Button)cp5.getController("Random")).setOff();
							((Button)cp5.getController("Direction")).setOff();
							
						} else {
							((Button)cp5.getController("Inertial")).setOn();
						}
					}
				});
		cp5Main.add(c);
		
		c = cp5.addButton("Direction")
				.setPosition(width-w,k2)
				.setSize(w,k)
				.setSwitch(true)
				.setCaptionLabel("Direction")
				.onClick(new CallbackListener() {
					public void controlEvent(CallbackEvent theEvent) {
						if ( ((Button)theEvent.getController()).getBooleanValue() ) {
							((Button)cp5.getController("Random")).setOff();
							((Button)cp5.getController("Inertial")).setOff();
							
						} else {
							((Button)cp5.getController("Direction")).setOn();
						}
					}
				});
		cp5Main.add(c);
		
		c = cp5.addButton("Best")
				.setPosition(width-w*3-2,k2+=10)
				.setSize(w,k)
				.setSwitch(true)
				.setCaptionLabel("Best")
				.onClick(new CallbackListener() {
					public void controlEvent(CallbackEvent theEvent) {
						if ( ((Button)theEvent.getController()).getBooleanValue() ) {
							((Button)cp5.getController("Rand_Works")).setOff();
							((Button)cp5.getController("Closest")).setOff();
						} else {
							((Button)cp5.getController("Best")).setOn();
						}
					}
				});
		cp5Main.add(c);
		
		c = cp5.addButton("Closest")
				.setPosition(width-w*2-1,k2)
				.setSize(w,k)
				.setSwitch(true)
				.setCaptionLabel("Closest")
				.onClick(new CallbackListener() {
					public void controlEvent(CallbackEvent theEvent) {
						if ( ((Button)theEvent.getController()).getBooleanValue() ) {
							((Button)cp5.getController("Rand_Works")).setOff();
							((Button)cp5.getController("Best")).setOff();
							
						} else {
							((Button)cp5.getController("Closest")).setOn();
						}
					}
				});
		cp5Main.add(c);
		
		c = cp5.addButton("Rand_Works")
				.setPosition(width-w,k2)
				.setSize(w,k)
				.setSwitch(true)
				.setOn()
				.setCaptionLabel("Random")
				.onClick(new CallbackListener() {
					public void controlEvent(CallbackEvent theEvent) {
						if ( ((Button)theEvent.getController()).getBooleanValue() ) {
							((Button)cp5.getController("Best")).setOff();
							((Button)cp5.getController("Closest")).setOff();
							
						} else {
							((Button)cp5.getController("Rand_Works")).setOn();
						}
					}
				});
		cp5Main.add(c);
		
		
		// global culture overrides
		c = cp5.addButton("Global_Override")
		.setPosition(width-w-200,k2+10)
		.setSize(80,k)
		.setSwitch(true)
		.setCaptionLabel("Global Override")
		.onClick(new CallbackListener() {
		      public void controlEvent(CallbackEvent theEvent) {
		    	  UpdateControlsVis();
		        }
		      });
		cp5Main.add(c);
		
		
		
		
		int sliX = width-w-tw-10-15-15;
		int texX = width-tw-10+5-15-15;
		int butX = width+5-15-15-15;
		int actX = width+5-15-15;
		
		for (int i = 0; i < 5; i++) {
			c = cp5.addButton("Culture "+i)
			.setPosition(width-w-110,k2+=10)
			.setColorBackground(colours[i])
			.setSize(w+110,k)
			.setCaptionLabel("Culture "+i)
			.setId(i)
			.setVisible( i < cp5.getController("Num_Cultures").getValue() );			
			cp5Main.add(c);

			
			AddCultureParameter (i, "Threshold", 0, 1, 0.4f, k2+=10, sliX, texX, butX, actX, w, k, tw );
			AddCultureParameter (i, "MaxThreshold", 0, 1, 1, k2+=10, sliX, texX, butX, actX, w, k, tw );
			AddCultureParameter (i, "Density", 0, 100, 75, k2+=10, sliX, texX, butX, actX, w, k, tw );
			AddCultureParameter (i, "Obligatoriness", 0, 1, 1, k2+=10, sliX, texX, butX, actX, w, k, tw );
			AddCultureParameter (i, "Exclusivity", 0, 1, 1, k2+=10, sliX, texX, butX, actX, w, k, tw );			
			AddCultureParameter (i, "NumVenues", 0, 10, 3, k2+=10, sliX, texX, butX, actX, w, k, tw );
			AddCultureParameter (i, "NeighDist", 0, 100, 4, k2+=10, sliX, texX, butX, actX, w, k, tw );
			AddCultureParameter (i, "VenDist", 0, 100, 4, k2+=10, sliX, texX, butX, actX, w, k, tw );
		}
		
		c = cp5.addButton("Run_Sim")
	     .setValue(0)
	     .setPosition(0,1)
	     .setSize(60,k)
	     ;
		cp5Main.add(c);

		cp5.addButton("Publish")
		.setValue(0)
		.setPosition(61,1)
		.setSize(40,k)
		.setCaptionLabel( "Publish" )
		.onPress(new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				PublishRun();
			}
		});
		
		
		
		
		
		
		cp5.addButton("Save_To_Server")
		.setValue(0)
		.setPosition(102,1)
		.setSize(40,k)
		.setCaptionLabel( "Save" )
		.onPress(new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				loadView = false;
				
				for (Controller cc : cp5Main)
					cc.setVisible(false);
				for (Controller cc : cp5Load)
					cc.setVisible(false);
				for (Controller cc : cp5Save)
					cc.setVisible(true);
				
				cp5.getController("Return").setVisible(true);
			}
		});

		cp5.addButton("Load_From_Server")
		.setValue(0)
		.setPosition(143,1)
		.setSize(40,k)
		.setCaptionLabel( "Load" )
		.onPress(new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				for (Controller cc : cp5Main)
					cc.setVisible(false);
				
				for (Controller cc : cp5Save)
					cc.setVisible(false);
				
				for (Controller cc : cp5Load)
					cc.setVisible(true);
				
				cp5.getController("Return").setVisible(true);
				
				loadStep = 0;
				loadView = true;
				
				fileList = srw.Load();
			}
		});
		
		c = cp5.addButton("Return")
		.setValue(0)
		.setPosition(184,1)
		.setSize(15,k)
		.setCaptionLabel( "R" )
		.onPress(new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				loadView = false;
				
				for (Controller cc : cp5Main) {
					cc.setVisible(true);
				}
				UpdateControlsVis();
				
				for (Controller cc : cp5Load)
					cc.setVisible(false);
				for (Controller cc : cp5Save)
					cc.setVisible(false);
			}
		});
		

		c = cp5.addButton("t")
	     .setValue(0)
	     .setPosition(240,1)
	     .setSize(39,k)
	     .setCaptionLabel( time + " / " + (maxTime-1) )
	     .onPress(new CallbackListener() {
		      public void controlEvent(CallbackEvent theEvent) {
		    	setTime(0);
		    	theEvent.getController().setCaptionLabel( time + " / " + (maxTime-1) );
		      }
	     });
		cp5Main.add(c);
		
		c = cp5.addButton(">")
	     .setValue(0)
	     .setPosition(280,1)
	     .setSize(19,k)
	     .onPress(new CallbackListener() {
		      public void controlEvent(CallbackEvent theEvent) {
		    	setTime(time+1);
		      }
	     });
		cp5Main.add(c);
		
		c = cp5.addButton(">>")
			     .setValue(0)
			     .setPosition(300,1)
			     .setSize(19,k)
			     .onPress(new CallbackListener() {
				      public void controlEvent(CallbackEvent theEvent) {
				    	setTime(maxTime);
				      }
			     });
				cp5Main.add(c);
		
		// put a label in the middle..
		c = cp5.addButton("<")
	     .setValue(0)
	     .setPosition(220,1)
	     .setSize(19,k)
	     .onPress(new CallbackListener() {
		      public void controlEvent(CallbackEvent theEvent) {
		    	setTime(time-1);
		      }
	     });
		cp5Main.add(c);
		

		
		
		/*  SAVE  */
		c = cp5.addTextfield("Author")
	    .setPosition(5,20)
	    .setSize(250,19)
	    .setText( "Author" );
		cp5Save.add(c);
		
		c = cp5.addTextfield("Description")
		.setPosition(5,40)
		.setSize(250,19)
		.setText( "Description" );
		cp5Save.add(c);
		
		c = cp5.addButton("Save_Upload")
		.setValue(0)
		.setPosition(5,60)
		.setSize(120,19)
		.setCaptionLabel( "Upload" )
		.onPress(new CallbackListener() {
			public void controlEvent(CallbackEvent theEvent) {
				srw.PermaSaveText(	((Textfield) cp5.getController("Author")).getText(), 
													((Textfield) cp5.getController("Description")).getText(),
													getGSettings(), getGAnalysis(sr),
													sr);
			}
		});
		cp5Save.add(c);
		
		
		for (Controller cc : cp5Save) {
			cc.setVisible(false);
		}

		for (Controller cc : cp5Load) {
			cc.setVisible(false);
		}
		
		/*  LOAD  */
		for (int i = 0; i < 3; i++) {
			c = cp5.addButton("Load_"+i)
					.setValue(0)
					.setPosition(width-240,50+ i * 150 + 135)
					.setSize(230,9)
					.setVisible(false)
					.setCaptionLabel( "Load" )
					.setId(i)
					.onPress(new CallbackListener() {
						public void controlEvent(CallbackEvent theEvent) {
							String fl = fileList[theEvent.getController().getId() + loadStep];
							LoadFromFile( fl.split(",")[0], fl.split(",")[1]);

							loadView = false;
							for (Controller cc : cp5Main) {
								cc.setVisible(true);
							}
							UpdateControlsVis();							
						}
					});
			cp5Load.add(c);
		}
	}

	public void HideLoadingUI() {
		if (!loadView) {
			for (Controller cc : cp5Load)
				cc.setVisible(false);
		}
	}
}
