package web_server.sim.schelling_180206;

import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.Collections;
import processing.core.PVector;
import processing.data.JSONArray;
import processing.data.JSONObject;
import web_server.sim.schelling_180206.Clusters.Cluster;

public class Sim_Context {
	public Agent[][] occupiedBy;
	
	public ArrayList<Agent_Human> a = new ArrayList<Agent_Human>();
	public ArrayList<Agent_Venue> v = new ArrayList<Agent_Venue>();
	int worldSize;
	int numCultures;
	boolean vonNeumann;
	int maxTime = 20;
	double randomMove = 0;
	String learningType;
	String movementType;
	
	public Sim_Context() {

	}
		
	/*
	 * This method is actually all about setting up for the run, it then returns a call to the method
	 * that actually does the work of going through the timesteps etc (SimulateAndRecord)
	 */
	public JSONObject run ( JSONObject jo ) {
		numCultures = jo.getInt("Num_Cultures");
		worldSize   = jo.getInt("World_Size");
		
		vonNeumann = jo.getBoolean("VN");
		maxTime = jo.getInt("Time_Steps");
		randomMove = jo.getDouble("Random_Moves");
		learningType = jo.getString("Learning");
		movementType = jo.getString("Movement");
		
		Culture[] cults = new Culture[numCultures];
		occupiedBy = new Agent[worldSize][worldSize];
		
		java.util.Random rndV = new java.util.Random();
		java.util.Random rndA = new java.util.Random();
		boolean lockedVenues = false;
		boolean lockedAgents = false;
		
		for (int i = 0; i < numCultures; i++) {
			Culture c = new Culture(jo.getJSONObject(i+"_settings"));
			if (c.VenueLocations.size() > 0) lockedVenues = true;		// some cultures might have 0 venues, but as long as 1 has more than 0, this will get tripped..
			if (c.AgentLocations.size() > 0) lockedAgents = true;
			cults[i] = c; 
		}

		if ( !lockedVenues ) {
			for (Culture c : cults) 
				for (int i = 0; i < c.numVenues; i++)
					v.add( new Agent_Venue( worldSize, numCultures, c, occupiedBy, rndV, null, vonNeumann, maxTime ) );
		} else {
			for (Culture c : cults) 
				for (PVector pv : c.VenueLocations)
					v.add(new Agent_Venue( worldSize, numCultures, c, occupiedBy, null, pv, vonNeumann, maxTime ) );
		}
		
		if (!lockedAgents) {
			for (Culture c : cults)
				for (int i = 0; i < (c.popDensity*worldSize*worldSize)/(float) numCultures; i++)
					a.add( new Agent_Human (worldSize, numCultures, c, movementType, v, occupiedBy, rndA, vonNeumann, learningType, maxTime, null, randomMove ));
		} else {
//			System.out.println("locked ");
			for (Culture c : cults) {
				for (int i = 0; i < c.AgentLocations.size(); i++)
					a.add( new Agent_Human (worldSize, numCultures, c, movementType, v, occupiedBy, rndA, vonNeumann, learningType, maxTime, c.AgentLocations.get(i), randomMove ));
			}
			for (Agent_Human haa : a) {
				if (haa.pos==null)
					haa.FindAPos(rndA);
			}
		}
		Collections.shuffle(a);
			
		return SimulateAndRecord();
	}
	
	JSONObject SimulateAndRecord() {
		JSONObject oneRun = new JSONObject();
		JSONObject analysis = new JSONObject();
		JSONObject results = new JSONObject();
		
		// a set of arrays storing some summary parameters for convenience,
		JSONArray sAvg = new JSONArray();			
		JSONArray sUnset = new JSONArray();  		
		JSONArray sMoved = new JSONArray();
		JSONArray sVenFrac = new JSONArray();		
		JSONArray jtime = new JSONArray();
		JSONArray jNumClusters = new JSONArray();
		JSONArray jAreaPerim = new JSONArray();
		JSONArray jMorphStability=  new JSONArray();
		
		int[][] lastMorphology = null;
		
		// main loop
		for (int time = 0; time < maxTime+1; time++) {
			// there are some thing that i want to calculate per time step - average ovrall diversity, etc. i should set up variables for those here
			int numUnset = 0;
			int numMoved = 0;
			float vFrac = 0;
			
			// shuffle every timestep
			Collections.shuffle(a);
			
			// every human should go to whatever venues they want to..
			// then we check for each of them whether they're happy - based on the venues they just visited
			// if they're not happy, they'll consider other spots, based on what's going on at the other venues this time step as well
			for (Agent_Human aa : a) {
				aa.GoToVenues( time );
			}
			
			for (Agent_Human aa : a) {
				aa.Update( time );
				
				if (aa.unset)
					numUnset++;
				if (aa.moved)
					numMoved++;
			}
			

			try {
			    Files.write(Paths.get ("moves.txt"), ("time " + time + "\n").getBytes(), StandardOpenOption.APPEND);
			}catch (Exception e) {

			}
			
			
			
			for (Agent_Venue vv : v) {
				vv.Update( time );
				vFrac += vv.getFractionLocalAgents( vv.travelDist );
			}
			if (v.size() > 0)
				vFrac /= (float) v.size();
			
			
			sAvg.append( getAverageDiversity() );
			sUnset.append(numUnset);
			sMoved.append(numMoved);
			sVenFrac.append(vFrac);
			
			// for now, the json is basically just a big comma separated string
			JSONObject jworld = getWorldJSON();
			jtime.append(jworld);
			
			int[][] morphology = generateMorphology();
			Clusters cl = new Clusters(morphology);
			ArrayList<Cluster> cc = cl.GenerateClusters();
			jNumClusters.append(cc.size());
			jAreaPerim.append(getPerimeterAreaRatio(cc));
			
			if (lastMorphology==null)
				jMorphStability.append(0);
			else
				jMorphStability.append( getMorphologicalDifference(morphology, lastMorphology));
			
			
			lastMorphology = morphology;
		}
		
		JSONObject storeAgents = new JSONObject();		// store both the agents and the venues
		JSONArray storeVenues = new JSONArray(), storeHumans = new JSONArray();
		
		for (Agent_Venue av : v)
			storeVenues.append( av.getJSON() );
		for (Agent_Human ah : a)
			storeHumans.append( ah.getJSON() );
		storeAgents.setJSONArray("Venues", storeVenues);
		storeAgents.setJSONArray("Humans", storeHumans);
		
		results.setJSONObject("Agents", storeAgents);
		results.setJSONArray("World", jtime);
		
		analysis.setJSONArray("qa_Average", sAvg);
		analysis.setJSONArray("qa_Unset", sUnset);
		analysis.setJSONArray("qa_Moved", sMoved);
		analysis.setJSONArray("qa_VenFrac", sVenFrac);
		analysis.setJSONArray("num_clusters", jNumClusters);
		analysis.setJSONArray("area_perim", jAreaPerim);
		analysis.setJSONArray("stability", jMorphStability);
		
		oneRun.setJSONObject("Results", results);
		oneRun.setJSONObject("Analysis", analysis);
		
		return oneRun;
	}
	
	// single step, moore neighbourhood
	public float getAverageDiversity () {
		float avgD = 0;
		int k = 0;

		for (int i = 0; i < occupiedBy.length; i++) {
			for (int j = 0; j < occupiedBy[0].length; j++) {

				if (occupiedBy[i][j] != null) {
					int c = 0; 					// neighbours like this
					int d = 0;					// total neighbours

					for (int a = -1; a < 2; a++) {
						for (int b = -1; b < 2; b++) {
							if ((a!=0||b!=0)&& i+a>-1 && i+a<occupiedBy.length && j+b>-1 && j+b<occupiedBy[0].length)  {

								if (occupiedBy[i+a][j+b] != null) {	
									d++;

									if ( occupiedBy[i][j].cult.identityCode == (occupiedBy[i+a][j+b].cult.identityCode)) {
										c++;
									}	
								}
							}
						}
					}
					if (d==0) {

					} else {
						float div = c/(float)d;
						avgD+=div;
						k++;
					}
				}
			}
		}
		return avgD/(float) k;		
	}
	
	public JSONObject getWorldJSON() {
		JSONObject wor = new JSONObject();
		String world = "";
		
		for (int i = 0; i < worldSize; i++) {
			for (int j = 0; j < worldSize; j++) {
				// write out a few things about this cell
				if (occupiedBy[i][j]!=null) {
					world += ( occupiedBy[i][j].type.charAt(0) + " " + occupiedBy[i][j].cult.identityCode + "," );
				} else {
					world += ("0,");
				}
			}
		}
		wor.setString("tempWorld", world);		
		return wor;
	}
	
	private float getPerimeterAreaRatio( ArrayList<Cluster> c) {
		float kz = 0;
		for (Cluster cc : c)
			kz += (cc.perim / (float) cc.area);
		kz /= (float) c.size();

		return kz;
	}
	
	private float getMorphologicalDifference(int[][] a, int[][] b) {
		int tot = 0;
		int dif = 0;
		
		for (int i = 0; i < a.length; i++) {
			for (int j = 0; j < a[i].length; j++) {
				if (a[i][j] != b[i][j]) {
					dif++;
				}
				tot++;
			}
		}
		return 1 - (dif / (float) tot);
	}	
	
	private int[][] generateMorphology () {
		int[][] wor = new int[worldSize][worldSize];
		
		for (int i = 0; i < wor.length; i++)
			for (int j = 0; j < wor[0].length; j++)
				wor[i][j] = (occupiedBy[i][j]!=null && occupiedBy[i][j].type.charAt(0) == 'h') ?  occupiedBy[i][j].cult.identityCode : -1;
		boolean finished = false;
		
		while (!finished) {
			finished = true;
			int[][] wort = new int[worldSize][worldSize];
			
			for (int i = 0; i < wor.length; i++) {
				for (int j = 0; j < wor[0].length; j++) {
					if (wor[i][j]>-1) {
						wort[i][j] = wor[i][j]; 
					} else {
						int fn = fromNeighbours(wor, i, j);
						wort[i][j] = fn;
						
						if (fn < 0)
							finished = false;
					}
				}
			}
			wor = wort;
		}
		return wor;
	}
	
	
	private int fromNeighbours(int[][] w, int i, int j) {
		// there must be a better way, but basically, get the most common culture of the neighbours.. if no neighbours return -1.
		int[] scores = new int[] { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1 };
		int hs = 0;
		
		for (int a = -1; a < 2; a++) {
			for (int b = -1; b < 2; b++) {
				if ( (a!=0 || b!= 0) && i+a > -1 && i+a < w.length && j+b > -1 && j+b < w[0].length && w[i+a][j+b] > -1 ) {
					scores[ w[i+a][j+b] ]++;
					if (scores[ w[i+a][j+b] ] > hs)
						hs = scores[ w[i+a][j+b] ];
				}
			}
		}

		if (hs > -1) {
			ArrayList<Integer> sc = new ArrayList<Integer>();
			for (int z = 0; z < scores.length; z++) {
				if (scores[z] == hs) {
					sc.add(z);
				}
			}

			if (sc.size() > 0)
				return sc.get( (int) (Math.random() * sc.size()) );
		}
		return -1;
	}
	
}
