package web_server.sim.schelling_180206;

import java.util.ArrayList;
import java.util.UUID;
import processing.data.JSONArray;
import processing.data.JSONObject;
import web_server.sim.schelling_180206.Clusters.Cluster;
import processing.core.PVector;
import processing.core.*;


/*
 * Takes the results from Sim_Context or from a txt file, and holds them for the sake of visualization and analysis
 * ie. takes a bunch of json and feeds out either a grid of values for the world or some summary values for charts
 * 
 */
public class SingleRun {
	
	// Results
	public String[][][] cells;
	ArrayList<Agent_Human> ha = new ArrayList<Agent_Human>();
	ArrayList<Agent_Venue> va = new ArrayList<Agent_Venue>();
	ArrayList<PVector> VenueLocations;
	
	// Analysis
	float[] DiversityPerTime;
	int[] UnsettledPerTime;
	int[] MovedPerTime;
	float[] VenFractionPerTime;
	int[] numClusters;
	float[] perimArea;
	float[] mStability;
	

	JSONObject jsP;				// parameters can be stored here, I don't think there's any reason to store the results in JSON form
	JSONObject jsR;				// since they are parsed into arrays if it's a thick rather than light version.. - - - if it's not saved yet?
	
	
	/*
	 * A single run takes in parameters (Settings), results (Raw), and analysis
	 * if it's a light-load, it will ignore the results and only load up parameters and analysis
	 * 
	 */
	public SingleRun (JSONObject jsP, JSONObject jsR, JSONObject jsA, boolean light ) {		
		this.jsP = jsP;
		
		// parse out some things
		UnsettledPerTime = jsA.getJSONArray("qa_Unset").getIntArray();
		MovedPerTime = jsA.getJSONArray("qa_Moved").getIntArray();
		VenFractionPerTime = jsA.getJSONArray("qa_VenFrac").getFloatArray();
		DiversityPerTime = jsA.getJSONArray("qa_Average").getFloatArray();
		numClusters = jsA.getJSONArray("num_clusters").getIntArray();
		perimArea = jsA.getJSONArray("area_perim").getFloatArray();
		mStability = jsA.getJSONArray("stability").getFloatArray();
	
		
		if (!light) {
			this.jsR = jsR;									// need this here, because right after generating, I tend to save to temp, but after that, it needs to be cleared
			
			int ws = jsP.getInt("World_Size");
			int ts = jsP.getInt("Time_Steps");		

			this.cells = new String[ts][][];
			VenueLocations = new ArrayList<PVector>();

			for (int i = 0; i < ts; i++) {
				JSONObject jow = jsR.getJSONArray("World").getJSONObject(i);
				String s2 = jow.getString("tempWorld");

				this.cells[i] = new String[ws][ws];

				String[] s3 = s2.split(",");

				for (int j = 0; j < s3.length; j++) {
					this.cells[i][j/ws][j%ws] = s3[j];

					if (i==0) {
						if (s3[j].contains("v")) {
							VenueLocations.add( new PVector (j/ws,j%ws, Integer.parseInt( s3[j].split(" ")[1] )));
						}
					}
				}
			}

			JSONArray jaH = jsR.getJSONObject("Agents").getJSONArray("Humans");
			for (int i = 0; i < jaH.size(); i++) {
				JSONObject haa = jaH.getJSONObject(i);

				Agent_Human ah = new Agent_Human(haa);
				//ah.cult = cul.get(ah.cultureIndex);

				ha.add( ah );
			}

			JSONArray jaV = jsR.getJSONObject("Agents").getJSONArray("Venues");
			for (int i = 0; i < jaV.size(); i++) {
				JSONObject vaa = jaV.getJSONObject(i);
				va.add( new Agent_Venue(vaa) );
			}
		}
	}
	
	
	// drop as much as possible out of memory, since the run has been saved to a temporary file
	public void LightenUp ( ) {
		// is it worth going through all of the above if it's just going to get cleared out here?
		cells = null;
		if (ha!= null) {
		ha.clear(); ha = null;}
		if (va!=null) {
		va.clear(); va = null;}
		if (VenueLocations!=null) {
		VenueLocations.clear(); VenueLocations = null; }
		
		jsR = null;
	}
	
	/* all of the get methods below are about re-interpreting the arrays so that they can be used to draw charts  */
	public ArrayList<PVector> getDiversityChart() {
		ArrayList<PVector> pv = new ArrayList<PVector>();

		for (int i = 0; i < DiversityPerTime.length; i++) {
			pv.add(new PVector(i, DiversityPerTime[i]));
		}			
		return pv;
	}

	public ArrayList<PVector> getUnhappinessChart() {
		ArrayList<PVector> pv = new ArrayList<PVector>();

		for (int i = 0; i < UnsettledPerTime.length; i++) {
			pv.add(new PVector(i, UnsettledPerTime[i]));
		}			
		return pv;
	}

	public ArrayList<PVector> getMovedChart() {
		ArrayList<PVector> pv = new ArrayList<PVector>();

		for (int i = 0; i < MovedPerTime.length; i++) {
			pv.add(new PVector(i, MovedPerTime[i]));
		}			
		return pv;
	}

	public ArrayList<PVector> getVenueNeighborhoods() {
		ArrayList<PVector> pv = new ArrayList<PVector>();

		for (int i = 0; i < VenFractionPerTime.length; i++) {
			pv.add(new PVector(i, VenFractionPerTime[i]));
		}			
		return pv;

	}

	public String getVenues() {
		String s = "";

		for (PVector pv : this.VenueLocations) {
			s += pv.x+" "+pv.y+" "+pv.z+",";
		}
		return s;
	}
	
	public ArrayList<PVector> getNumClusters() {
		ArrayList<PVector> pv = new ArrayList<PVector>();

		for (int i = 0; i < numClusters.length; i++) {
			pv.add(new PVector(i, numClusters[i]));
		}			
		return pv;

	}
	
	public ArrayList<PVector> getPerimArea() {
		ArrayList<PVector> pv = new ArrayList<PVector>();

		for (int i = 0; i < perimArea.length; i++) {
			pv.add(new PVector(i, perimArea[i]));
		}			
		return pv;

	}
	
	public ArrayList<PVector> getStability() {
		ArrayList<PVector> pv = new ArrayList<PVector>();

		for (int i = 0; i < mStability.length; i++) {
			pv.add(new PVector(i, mStability[i]));
		}			
		return pv;
	}
}
